/*
 * cdcc.c: dcc client offers. 
 *
 * Written By Colten Edwards <edwac@sasknet.sk.ca> 
 *
 * Copyright(c) 1996
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */
#define QUEUE_DCC
#define TEST_QUEUE

#include "irc.h"

#ifdef ESIX
# include <lan/net_types.h>
#endif /* ESIX */

#if defined(ISC30) && defined(_POSIX_SOURCE)
# undef _POSIX_SOURCE
#include <sys/stat.h>
# define _POSIX_SOURCE
#else
# include <sys/stat.h>
#endif /* ICS30 || _POSIX_SOURCE */

#include <dirent.h>
#include <sys/stat.h>
#include "talkd.h"
#include "server.h"
#include "ircaux.h"
#include "whois.h"
#include "lastlog.h"
#include "ctcp.h"
#include "dcc.h"
#include "hook.h"
#include "vars.h"
#include "window.h"
#include "output.h"
#include "newio.h"
#include "list.h"
#include "timer.h"
#include "screen.h"
#include "pluto.h"

char *	dcc_addpack _((char *, char *, char *));
char *	dcc_delpack _((char *, char *, char *));
char *	dcc_listpack _((char *, char *, char *));
char *	dcc_listp _((char *, char *, char *));
char *	dcc_offer _((char *, char *, char *));
char *	dcc_notice _((char *, char *, char *));
char *	dcc_help _((char *, char *, char *));
char *	dcc_mhelp _((char *, char *, char *));
char *	dcc_showme _((char *, char *, char *));
char *	dcc_getpack _((char *, char *, char *));
char *	dcc_sendpack _((char *, char *, char *));
char *	dcc_show_queue _((char *, char *, char *));
char *	dcc_timer _((char *, char *, char *));

extern int dcc_active_count;
extern void dcc_filesend _((char *));
extern void dcc_show_active _((char *));

FileList *QueueList = NULL;

struct 
{
	char * name;
	int user;
	char * (*func) _((char *, char *, char *));
} cmdslist[2][12] =
{{
	{ "LIST",	1, dcc_listpack }, /* User Commands */
	{ "PLIST",	0, dcc_offer },
	{ "OFFER",	0, dcc_addpack },
	{ "GET",	0, dcc_getpack },
	{ "SEND",	1, dcc_sendpack },
	{ "RESEND",	1, dcc_sendpack },
	{ "DOFFER",	0, dcc_delpack },
	{ "NOTICE",	0, dcc_notice },
	{ "HELP",	1, dcc_mhelp },
#ifdef QUEUE_DCC
	{ "QUEUE",	0, dcc_show_queue },
#endif
	{ "TIMER",	1, dcc_timer },
	{ "",		0, NULL},
},{
	{ "LIST",	0, dcc_listp }, /* Local Commands */
	{ "PLIST",	0, dcc_offer },
	{ "OFFER",	0, dcc_addpack },
	{ "GET",	0, dcc_getpack },
	{ "SEND",	0, dcc_sendpack },
	{ "RESEND",	0, dcc_sendpack },
	{ "DOFFER",	0, dcc_delpack },
	{ "NOTICE",	0, dcc_notice },
	{ "HELP",	0, dcc_help },
#ifdef QUEUE_DCC
	{ "QUEUE",	0, dcc_show_queue },
#endif
	{ "TIMER",	0, dcc_timer },
	{ "",		0, NULL},
}
};

FileList *file_lists = NULL;
FileList *new = NULL;

static char pack_description[1000];
static int  packnumber = 0;
static int dcc_timer_val = 0;
#define DEFAULT_TIMER_VAL 60

int number_in_queue = 0;

char *msgcdcc(char *from, char *to, char *args ) 
{
int i; 
char *p = NULL, *q;
char *ptr = args;
char *xdcc, *comm;
	if (!file_lists)
		return args;
	malloc_strcpy(&p, args);
	q = p;
	xdcc = next_arg(p, &p);
	if (!(comm = next_arg(p, &p)))
	{
		new_free(&q);
		return args;
	}
	if ((my_strnicmp(args, "XDCC",4) == 0) || (my_strnicmp(args,"CDCC",4) == 0) || (my_strnicmp(args, "BDCC",4) == 0)) {
		for (i = 0; *cmdslist[0][i].name; i++ ) {
			if ((my_stricmp(comm, cmdslist[0][i].name) == 0) && (cmdslist[0][i].func != NULL) && cmdslist[0][i].user){
				if (!((*cmdslist[0][i].func)(p, from, NULL)))
					ptr = NULL;
				break;
				
			}
		}
		if (ptr)
			send_to_server("NOTICE %s :\002Unknown CDCC command error. Try /msg %s CDCC help\002", from, to);
	}
	new_free(&q);
	return ptr;
}

void cdcc(char *command, char *args, char *subargs ) 
{
int i; char *from, *to;
	from = to = NULL;
	if (!args || !*args) {
		bitchsay("Try /CDCC help");
		return;
	}
	from  = next_arg(args, &args);
	for (i = 0; *cmdslist[1][i].name; i++ ) {
		if ((my_stricmp(from, cmdslist[1][i].name) == 0) && (cmdslist[1][i].func != NULL) && !cmdslist[1][i].user){
			(*cmdslist[1][i].func)(args, from, to);
			break;
		}
	}
}

char *do_bdcc(CtcpEntry *ctcp, char *from, char *to, char *args )
{
char *p;
int i;
int channel_ctcp = 0;
char *ptr = args;
	if (!args || !*args)
		return NULL;
	p  = next_arg(args, &args);
	if (*to == '#')
		channel_ctcp++;
	for (i = 0; *cmdslist[0][i].name; i++ ) {
		if ((my_stricmp(p, cmdslist[0][i].name) == 0) && (cmdslist[0][i].func != NULL) && cmdslist[0][i].user){
			(*cmdslist[0][i].func)(args, from, to); 
			ptr = NULL;
			break;
			
		}
		ptr = NULL;
	}
	return ptr;
}
#ifdef QUEUE_DCC
char *dcc_show_queue(char *args, char *from, char *to)
{
FileList *tmp = QueueList;
	if (tmp) {
		if (do_hook(CDCC_QUEUEH_LIST, "%s %s %s", "Nick","Time","Description"))
			put_it(" %-10s %-25s %s", "Nick", "Time", "Description");
		for (tmp = QueueList; tmp; tmp = tmp->next)
			if (do_hook(CDCC_QUEUE_LIST, "%s %s %s", tmp->nick, my_ctime(tmp->timequeue), tmp->description))
				put_it(" %-10s %-25s %s ", tmp->nick, my_ctime((tmp->timequeue)), tmp->description); 
	} else
		bitchsay("No Files Queued for send");
	return NULL;
}


void send_file_from_queue(char *unused, char *nicklist)
{
FileList *tmp = NULL;
char * p = NULL;
	/* take the top entry in the list and send it */
	if ((tmp = QueueList))
	{
		QueueList = tmp->next;
		tmp->next = NULL;
	} else 
	{
		bitchsay("ERROR in dcc send from queue");
		return;
	}
	bitchsay("Queue File Send to %s", tmp->nick, tmp->description);
	if (nicklist && *nicklist)
	{
		malloc_sprintf(&p, "%s %s", tmp->nick, tmp->filename->filename);
		dcc_filesend(p);
		new_free(&p);
	}
	new_free(&tmp->nick);
	new_free(&tmp->description);
	new_free(&tmp->filename->filename);
	new_free((char **)&tmp->filename);
	new_free((char **)&tmp);
	if (number_in_queue) number_in_queue--;
}

void dcc_sendfrom_queue(void)
{
	if (QueueList && (dcc_active_count < get_int_var(DCC_SEND_LIMIT_VAR))) 
		add_ison_to_whois(QueueList->nick, send_file_from_queue);
}

void dcc_queue(char *nick, char *filename, char *description)
{
FileList *tmp = NULL, *tmp1 = NULL;
	tmp = (FileList *) new_malloc(sizeof(FileList));
	memset(tmp, 0, sizeof(FileList));
	malloc_strcpy(&tmp->nick, nick);
	tmp->filename = (FileStat *)new_malloc(sizeof(FileStat));
	malloc_strcpy(&tmp->filename->filename, filename);
	malloc_strcpy(&tmp->description, description);
	tmp->timequeue = time(NULL);
	number_in_queue++;
	send_to_server("NOTICE %s :\002File Sends are currently busy. File %s has been Queued\002", nick, description); 
	send_to_server("NOTICE %s :\002I will send as soon as a spot on the queue is open. You are %d on queue.\002", nick, number_in_queue); 
	if (QueueList)
	{
		/* find the end of the list and add the new entry */
		for (tmp1 = QueueList; tmp1; tmp1 = tmp1->next)
		{
			if (tmp1->next == NULL)
			{
				tmp1->next = tmp;
				break;
			}
		}
	
	}
	else
		QueueList = tmp;
}
#endif

char *dcc_sendpack (char *args, char *from, char *to)
{
static int lame_counter = 0;
char *channel, *p, *nick = NULL;
int file_no = 0;
FileList *files = file_lists;

	p = next_arg(args, &args);
	if (p && *p)
	{
		if (*p == '#')
			p++;
		file_no = atoi(p);
		if (file_no  <= 0)
			return from;
	} else
		return from;
	if ((nick = next_arg(args, &args)) == NULL)
		nick = from;
	for (files = file_lists; files != NULL; files = files->next) {
		if (files->packnumber == file_no) {
			files->gets++;
			break;
		}
	}
	channel = get_channel_by_refnum(0);
	if (files && get_int_var(AUTOCK_VAR) && (channel != NULL) && !find_in_list((List **)&user_list, from, 0)) {
		if (you_is_opped(channel))
			send_to_server("KICK %s %s :\002LaMe CTCP KiCk CoUnT #%d\002", channel, from, ++lame_counter);
		return from;
	}
	if (files != NULL) {
		FileStat *t;
		p = NULL;
		for (t = files->filename; t; t = t->next)
		{
			if (get_int_var(DCC_SEND_LIMIT_VAR) && (dcc_active_count >= get_int_var(DCC_SEND_LIMIT_VAR)))
			{
				dcc_queue(nick, t->filename, files->description);
				continue;
			}
			malloc_sprintf(&p, "%s %s", nick, t->filename);
			dcc_filesend(p);
			new_free(&p);
		}
		return NULL;
	}
	return from;
}

char *dcc_getpack (char *args, char *from, char *to)
{
/*	nick = next_arg(args, &args);*/
	return NULL;
}

char *dcc_listpack (char *args, char *from, char *to) 
{
FileList *files = file_lists;
	if (files != NULL) {
		if (do_hook(CDCC_PREPACK_LIST, "%s %s %d %d %d", from, nickname, packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR)))
		{
			send_to_server("NOTICE %s :Files Offered: /ctcp %s CDCC send #N for pack N", from, nickname);
			if (get_int_var(DCC_SEND_LIMIT_VAR))
				send_to_server("NOTICE %s :    [%d packs %d/%d slots open]", from, 
					packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
			else
				send_to_server("NOTICE %s :    [%d packs]", from, packnumber);
		}
		for (files = file_lists; files != NULL; files = files->next) {
			if (do_hook(CDCC_PACK_LIST, "%s %d %d %2.3f %d %s", 
				from, files->packnumber, 
				files->numberfiles, files->filesize, 
				files->gets, files->description))
			{
				send_to_server("NOTICE %s :\002#%d\002:  %s [\002%d\002 File%s/\002%-2.3f\002 Kb] %d gets", from, 
				files->packnumber, files->description, 
				files->numberfiles, files->numberfiles == 1 ? "":"s",
				files->filesize/(double)1000, files->gets);
			}
		}	
		(void) do_hook(CDCC_POSTPACK_LIST, "%s %s %d %d %d", from, nickname, packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
	} 
	return NULL;
}

char *dcc_listp (char *args, char *from, char *to) 
{
FileList *files = file_lists;
	if (files != NULL) {
		for (files = file_lists; files != NULL; files = files->next) {
			put_it(":\002#%d\002:  %s [\002%d\002 File%s/\002%-2.3f\002 Kb] %d gets",
			files->packnumber, files->description, 
			files->numberfiles, files->numberfiles == 1 ? "":"s",
			files->filesize/(double)1000, files->gets);
		}	
	} else bitchsay("No Files Offered");
	return NULL;
}

void add_desc(char *args, char *subargs)
{
FileList *files = NULL;
int renumber;
	strcpy(pack_description, subargs);
	if (new) {
		malloc_strcpy(&new->description, pack_description);
		add_to_list((List **) &file_lists, (List *)new);
		bitchsay("Created new pack : [\002%d\002 File%s/%-2.3f Kb]", new->numberfiles,
			new->numberfiles == 1? "":"s", new->filesize/(double)1000);
		for (renumber = 1, files = file_lists; files != NULL; files= files->next, renumber++)
			files->packnumber = renumber;
	}
}

void add_file(char *args, char *subargs) 
{
char path[BIG_BUFFER_SIZE+1];
char filebuf[BIG_BUFFER_SIZE+1];
char *thefile, *expand;
DIR *dp;
struct dirent *dir;
struct stat stbuf;
static int been_here = 0;
FileStat *t;

	if (been_here == 0) {
		been_here ++;
		file_lists = NULL;
	}
	if (new) new = NULL;	
	strcpy(path, subargs);
        if ((thefile = strrchr(path, '/')) != NULL)
              *thefile++ = '\0';
        else {
              strcpy(path, "~");
              thefile = subargs;
        }
        expand = expand_twiddle(path);
	dp = opendir(expand);
	if (!dp) {
		bitchsay("You cannot access dir %s", expand);
		new_free(&expand);
		return;
	}

	bitchsay("Searching %s/%s", expand, thefile);
        new = (FileList *)new_malloc(sizeof(FileList));
        memset(new, 0, sizeof(FileList));

	while ((dir = readdir(dp)) != NULL) {
                if (dir->d_ino == 0)
                       continue;
                if (!wild_match(thefile, dir->d_name))
                       continue;
                sprintf(filebuf, "%s/%s", expand, dir->d_name);
                stat(filebuf, &stbuf);
                if (S_ISDIR(stbuf.st_mode))
                       continue;
                new->filesize += stbuf.st_size;
                new->numberfiles++;
		t = (FileStat *)new_malloc(sizeof(FileStat));
		malloc_sprintf(&t->filename, "%s/%s", expand, dir->d_name);
		t->filesize = stbuf.st_size;
		t->next = new->filename;
		new->filename = t;
	}

	if (new->numberfiles) {
                new->packnumber = ++packnumber;
		bitchsay("Added %d files to Pack....", new->numberfiles);
		add_wait_prompt("Description of Pack? ", add_desc, "", WAIT_PROMPT_LINE);
	} else { 
		bitchsay("No Files for Pack found, Aborting....");
		if (new) new_free((char **)&new);
		new = NULL;
	}
	new_free(&expand);
}

char *dcc_addpack (char *args, char *from, char *to)  
{
	add_wait_prompt("Add Files to Pack? ", add_file, "", WAIT_PROMPT_LINE);
	return NULL;
}

void clear_files(FileStat **filenames)
{
FileStat *t, *t1;
	for (t = *filenames; t;)
	{
		new_free(&t->filename);
		t1 = t->next;
		new_free((char **)&t);
		t = t1;
	}
	*filenames = NULL;
}

void clear_offer(FileList **clientlist)
{
        FileList        *Client, *tmp;

	for (Client = *clientlist; Client;)
	{
		clear_files(&Client->filename);
		new_free(&Client->description);
		new_free(&Client->nick);
		tmp = Client->next;
		new_free((char **)&Client);
		Client = tmp;
	}
	*clientlist = NULL;
}

void del_file(char *args, char *subargs) 
{
FileList *files = file_lists;
FileList *delpack = NULL;
int packno = 1;
int renumber;

	if (*subargs != '*')
	{
		packno = atoi(subargs);
		if (!packno) {
			say("Need to number between 1 and last pack number");
			return;
		}
	} 
	else
	{
		clear_offer(&file_lists);
		packnumber = 0;
		return;
	}
	if (files != NULL) {
		char *tmpname = NULL;
		for (files = file_lists; files != NULL; files = files->next) {
			if (files->packnumber == packno) {
				tmpname = files->description;
				break;
			}
		}
		if (tmpname && (delpack = (FileList *)remove_from_list((List **)&file_lists, tmpname)) != NULL) {
			clear_files(&delpack->filename);
			new_free(&delpack->description);
			new_free((char **)&delpack);
			delpack = NULL;
		}
		for (packnumber = 0, renumber = 1, files = file_lists; files != NULL; files = files->next, renumber++, packnumber++) 
			files->packnumber = renumber;
	} else say("No Packs to delete");
}

char *dcc_delpack (char *args, char *from, char *to)
{
FileList *files;
	if (file_lists == NULL)
	{
		bitchsay("No Files Offered");
		return NULL;
	}
	for (files = file_lists; files; files = files->next)
		put_it("#%2d %s", files->packnumber, files->description);
	add_wait_prompt("Delete Pack [* for all]? ", del_file, "", WAIT_PROMPT_LINE);
	return NULL;
}

char *dcc_offer (char *args, char *from, char *to)
{
FileList *files = file_lists;
char *channel = NULL;
char what[BIG_BUFFER_SIZE + 1];
	if (files == NULL)
	{
		bitchsay("No Files Offered");
		return NULL;
	}
	if (get_channel_by_refnum(0)) {
		malloc_strcpy(&channel, get_channel_by_refnum(0));
		if (do_hook(CDCC_PREPACK_LIST, "%s %s %d %d %d", channel, nickname, packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR)))
		{

			sprintf(what, "Files Offered: /Ctcp %s \002CDCC\002 list", nickname);
			send_text(channel, what, NULL);
			if (get_int_var(DCC_SEND_LIMIT_VAR))
				sprintf(what, "    [%d packs %d/%d slots open]",packnumber, 
					get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
			else
				sprintf(what, "    [%d packs]",packnumber); 
			send_text(channel, what, NULL);
		}
		for (files = file_lists; files != NULL; files = files->next) 
		{
			if (do_hook(CDCC_PACK_LIST, "%s %d %d %2.3f %d %s", 
				channel, files->packnumber, files->numberfiles, 
				files->filesize, files->gets, files->description))
			{
				sprintf(what, "\002#%d\002:  %s [\002%d\002 File%s/\002%-2.3f\002 Kb] %d gets",
				files->packnumber, files->description ? files->description: "No Description", 
				files->numberfiles, files->numberfiles == 1 ? "":"s",
				files->filesize/(double)1000, files->gets);
				send_text(channel,what, NULL);
			}
		}	
		(void) do_hook(CDCC_POSTPACK_LIST, "%s %s %d %d %d", from, nickname, packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
		new_free(&channel);
	}
	return NULL;
}

char *dcc_notice (char *args, char *from, char *to)
{
char *channel = NULL;
char *msg = NULL;

	if (get_channel_by_refnum(0)) {
		if (file_lists)
		{
			malloc_strcpy(&channel, get_channel_by_refnum(0));
			if (do_hook(CDCC_PREPACK_LIST, "%s %s %d %d %d", channel, nickname, packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR)))
			{
			
				malloc_sprintf(&msg, "\002== Files Offered\002  /Ctcp %s \002CDCC\002 list", nickname);
				send_text(channel, msg, NULL);
				if (get_int_var(DCC_SEND_LIMIT_VAR))
					malloc_sprintf(&msg, "    [%d packs %d/%d slots open]",
						packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
				else
					malloc_sprintf(&msg, "    [%d packs]",packnumber);
			
				send_text(channel, msg, NULL);	
			}
			(void) do_hook(CDCC_POSTPACK_LIST, "%s %s %d %d %d", from, nickname, packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
			new_free(&channel);
			new_free(&msg);
		}
		else
			bitchsay("No Files to offer");
	}
	return NULL;
}

int do_plist_timer(void *args)
{
	char *p = (char *)args;
	
	char *timer, *channel, *q = NULL;
	char what[BIG_BUFFER_SIZE+1];
	FileList *files = file_lists;
	
	malloc_strcpy(&q, p);
	timer = next_arg(q, &q);
	channel = next_arg(q, &q);	

	if (dcc_timer_val == 0 || !im_on_channel(channel) || !files)
	{
		new_free(&p);
		new_free(&q);
		return 0;
	}
	
	add_timer(-1, dcc_timer_val, do_plist_timer, (char *)args, NULL);
	if (do_hook(CDCC_PREPACK_LIST, "%s %s %d %d %d", channel, nickname, packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR)))
	{

		sprintf(what, "Files Offered: /Ctcp %s \002CDCC\002 list", nickname);
		send_text(channel, what, NULL);
		if (get_int_var(DCC_SEND_LIMIT_VAR))
			sprintf(what, "    [%d packs %d/%d slots open]",packnumber, 
				get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
		else
			sprintf(what, "    [%d packs]",packnumber); 
		send_text(channel, what, NULL);
	}

	for (files = file_lists; files != NULL; files = files->next) 
	{
		if (do_hook(CDCC_PACK_LIST, "%s %d %d %2.3f %d %s", 
			channel, files->packnumber, files->numberfiles, 
			files->filesize, files->gets, files->description))
		{
			sprintf(what, "\002#%d\002:  %s [\002%d\002 File%s/\002%-2.3f\002 Kb] %d gets",
			files->packnumber, files->description ? files->description: "No Description", 
			files->numberfiles, files->numberfiles == 1 ? "":"s",
			files->filesize/(double)1000, files->gets);
			send_text(channel,what, NULL);
		}
	}	
	(void) do_hook(CDCC_POSTPACK_LIST, "%s %s %d %d %d", channel, nickname, packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
	new_free(&timer);
	return 0;
}

void add_plist_timer(char *args, char *subargs)
{
char *p = NULL;
char *timer, *channel;
	if (!subargs && !*subargs)
	{
		dcc_timer_val = 0;
		return;
	}
	if (*subargs == '-')
	{
		bitchsay("Removing timer plist");
		dcc_timer_val = 0;
		return;
	}
	timer = next_arg(subargs, &subargs);
	dcc_timer_val = atoi(timer);
	if (dcc_timer_val < DEFAULT_TIMER_VAL)
		dcc_timer_val = DEFAULT_TIMER_VAL;

	if (args && *args)
		channel = next_arg(subargs, &subargs);
	else
		channel = get_channel_by_refnum(0);

	if (!channel)
	{
		bitchsay("No Channel");
		return;		
	}
	bitchsay("Timer set to %d secs for %s", dcc_timer_val, channel);
	malloc_sprintf(&p, "%d %s", dcc_timer_val, channel);
	add_timer(-1, dcc_timer_val, do_plist_timer, p, NULL);
}

char *dcc_timer(char *args, char *from, char *to)
{
	add_wait_prompt("Delay for plist? [- to delete]", add_plist_timer, "", WAIT_PROMPT_LINE);
	return NULL;
}

char *dcc_help (char *args, char *from, char *to)
{
	put_it("\002CDCC\002: get	- pack from a user");
	put_it("\002CDCC\002: send 	- send a pack to nick");
	put_it("\002CDCC\002: offer	- add a pack");
	put_it("\002CDCC\002: doffer	- delete a pack");
	put_it("\002CDCC\002: list	- list pack for you");
	put_it("\002CDCC\002: plist	- public list a pack");
	put_it("\002CDCC\002: notice	- announce to channel a pack");
	put_it("\002CDCC\002: queue	- show queue");
	put_it("\002CDCC\002: timer	- plist timer");
	return NULL;
}

char *dcc_mhelp (char *args, char *from, char *to)
{
	send_to_server("NOTICE %s :\002 SEND #  - send a file pack", from);
	send_to_server("NOTICE %s :\002 LIST    - list file packs", from);
	send_to_server("NOTICE %s :\002 HELP    - This help msg", from);
	return NULL;
}
