/*
 * dcc.c: Things dealing client to client connections. 
 *
 * Written By Troy Rollo <troy@cbme.unsw.oz.au> 
 *
 * Copyright(c) 1991 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#if 0
static	char	rcsid[] = "@(#)$Id: dcc.c,v 1.37 1994/07/26 15:33:23 mrg Exp $";
#endif

#include "irc.h"

#ifdef ESIX
# include <lan/net_types.h>
#endif /* ESIX */

#if defined(ISC30) && defined(_POSIX_SOURCE)
# undef _POSIX_SOURCE
#include <sys/stat.h>
# define _POSIX_SOURCE
#else
# include <sys/stat.h>
#endif /* ICS30 || _POSIX_SOURCE */

#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#include <stdio.h>

#include <dirent.h>

#include "ctcp.h"
#include "dcc.h"
#include "hook.h"
#include "ircaux.h"
#include "lastlog.h"
#include "newio.h"
#include "output.h"
#include "parse.h"
#include "server.h"
#include "status.h"
#include "vars.h"
#include "whois.h"
#include "window.h"
#include "ircterm.h"
#include "hook.h"
#include <float.h>

static	off_t		filesize = 0;
static	DCC_list *	ClientList = NULL;
static	char		DCC_current_transfer_buffer[256];
static	int		DCC_done = 1;
extern	int		dgets_errno;
extern	int		in_ctcp_flag;
extern	char	MyHostName[];
/*extern	struct	in_addr	MyHostAddr;*/


static	void    	add_to_dcc_buffer _((DCC_list *, char *));
static	void 		dcc_add_deadclient _((DCC_list *));
	void		dcc_chat _((char *));
static	void 		dcc_close _((char *));
static	void		dcc_erase _((DCC_list *));
	void		dcc_filesend _((char *));
	void		dcc_list _((char *));
	void		dcc_glist _((char *));
static	void		dcc_getfile _((char *));
static	int		dcc_open _((DCC_list *));
static	void 		dcc_really_erase _((void));
static	void		dcc_rename _((char *));
static	DCC_list *	dcc_searchlist _((char *, char *, int, int, char *, int));
static	void		dcc_send_raw _((char *));
static	char *		dcc_time _((time_t));
static	void 		output_reject_ctcp _((char *, char *));
static	void		process_incoming_chat _((DCC_list *));
static	void		process_incoming_listen _((DCC_list *));
static	void		process_incoming_raw _((DCC_list *));
static	void		process_outgoing_file _((DCC_list *));
static	void		process_incoming_file _((DCC_list *));
static	void		file_checksum _((unsigned char foo[], char *));
static	void		DCC_close_filesend _((DCC_list *));
static	void		update_transfer_buffer _((char *format, ...));
#ifdef NON_BLOCKING_CONNECTS
static	void	dcc_got_connected _((DCC_list *));
#endif
static  void    dcc_stats _((char *));
static  void    dcc_update_stats _((DCC_list *));
static	void	dcc_set_paths _((char *));
static  void	dcc_set_quiet _((char *));
static  void	dcc_tog_auto _((char *));
static	void	dcc_show_active _((char *));
static	void	dcc_help1 _((char *));
void		multiget _((char *, char *));
void		multisend _((char *, char*));
void		dcc_resend _((char *));
void		dcc_regetfile _((char *));
static unsigned char byteordertest _((void));
static void dcc_reject_notify _((char *, char *, char *));
static	void 	output_reject_ctcp (char *notused, char *nicklist);
static	int	get_to_from _((char *));

static char * strip_path _((char *));
static void dcc_set_quiet _(( char *)); 
static void dcc_show_active _(( char *)); 
static void dcc_help1 _((char *));
static char *check_paths _((char *));
static void dcc_set_paths _((char *));
static void dcc_overwrite_toggle _((char *));
void dcc_tog_auto _((char *));
static void dcc_stats _((char *));
static void dcc_update_stats _((DCC_list *));

typedef void (*dcc_function) _((char *));
struct
{
	char	*	name;
	dcc_function 	function;
}	dcc_commands[] =
{
	{ "ACTIVE",	dcc_show_active },
	{ "AUTO",	dcc_tog_auto },
	{ "CHAT",	dcc_chat },
	{ "LIST",	dcc_list },
	{ "GLIST",	dcc_glist },
	{ "SEND",	dcc_filesend },
	{ "RESEND",	dcc_resend }, 
	{ "GET",	dcc_getfile },
	{ "REGET",	dcc_regetfile },
	{ "CLOSE",	dcc_close },
	{ "RENAME",	dcc_rename },
	{ "RAW",	dcc_send_raw },
	{ "QUIET",	dcc_set_quiet },
	{ "RENAME",	dcc_rename },
	{ "RAW",	dcc_send_raw },
	{ "STATS",	dcc_stats },
	{ "PATHS",	dcc_set_paths },
	{ "OVERWRITE",	dcc_overwrite_toggle },
	{ "?",		dcc_help1 },
	{ NULL,		(dcc_function) NULL }
};

#define BAR_LENGTH 50
static  int	dcc_count = 1;
	int	dcc_active_count = 0;
static	int	doing_multi = 0;
static	int	dcc_quiet = 0;
static	int	dcc_paths = 0;
static	int	dcc_overwrite_var = 0;
static unsigned long dcc_bytes_in = 0;
static unsigned long dcc_bytes_out = 0;
static double   dcc_max_rate_in = 0.0;
static double   dcc_min_rate_in = DBL_MAX;
static double   dcc_max_rate_out = 0.0;
static double   dcc_min_rate_out = DBL_MAX;
static int      dcc_count_stat = 0;


	char	*dcc_types[] =
{
	"<none>",
	"CHAT",
	"SEND",
	"GET",
	"RAW_LISTEN",
	"RAW",
	"RESEND",
	"REGET",
	NULL
};

struct	deadlist
{
	DCC_list *it;
	struct deadlist *next;
}	*deadlist = NULL;


/*
 * dcc_searchlist searches through the dcc_list and finds the client
 * with the the flag described in type set.
 */
static
#ifdef __STDC__
DCC_list *dcc_searchlist(char *name, char *user, int type, int flag, char *othername, int active)
#else
DCC_list * dcc_searchlist(name, user, type, flag, othername, active)
	char	*name,
		*user;
	int	type,
		flag;
	char	*othername;
	int	active;
#endif
{
	DCC_list **Client, *NewClient;

#ifdef DEBUG
	yell("entering dcc_sl.  name (%s) user (%s) other (%s)", name, user, othername);
#endif
	int dcc_num = 0;
	if (user && *user == '#' && atoi(user+1)) 
	{
		/* we got a number so find that instead */
		char *p;
		p = user;
		p++;
		dcc_num = atoi(p);
		if (dcc_num > 0)
		{
			for(Client = (&ClientList); *Client; Client = (&(**Client).next))
				if (dcc_num && ((**Client).dccnum == dcc_num))
					return *Client;
		}
	} 
	else
	{
	for (Client = (&ClientList); *Client ; Client = (&(**Client).next))
	{
		/* 
		 * The following things have to be true:
		 * -> The types have to be the same
		 * -> One of the following things is true:
		 *      -> `name' is NULL or `name' is the same as the 
		 *	   entry's description
		 *          *OR*
		 *	-> `othername' is the same as the entry's othername
		 * -> `user' is the same as the entry's user-perr
		 * -> One of the following is true:
		 *	-> `active' is 1 and the entry is active
		 *	-> `active' is 0 and the entry is not active
		 *	-> `active' is -1 (dont care)
		 */
		if (
		     ( ((**Client).flags & DCC_TYPES) == type ) &&
		     ( (!name || ((**Client).description && !my_stricmp(name, (**Client).description))) ||
		       (othername && (**Client).othername && !my_stricmp(othername, (**Client).othername))   ) &&
		     ( !my_stricmp(user, (**Client).user) ) &&
		     (  (active == -1) ||
		       ((active == 0) && !((**Client).flags & DCC_ACTIVE)) ||
		       ((active == 1) && ((**Client).flags & DCC_ACTIVE))
		     )
		   )
			return *Client;
	}
	}
	if (!flag)
		return NULL;
	*Client = NewClient = (DCC_list *) new_malloc(sizeof(DCC_list));
	NewClient->flags = type;
	NewClient->read = NewClient->write = NewClient->file = -1;
	NewClient->filesize = filesize;
	NewClient->next = NULL;
	NewClient->user = NewClient->description = NewClient->othername = NULL;
	NewClient->bytes_read = NewClient->bytes_sent = 0L;
	NewClient->starttime = 0;
	NewClient->buffer = 0;
	NewClient->window_max = 0;
	NewClient->window_sent = 0;
	NewClient->cksum = NULL;
	malloc_strcpy(&NewClient->description, name);
	malloc_strcpy(&NewClient->user, user);
	malloc_strcpy(&NewClient->othername, othername);
	NewClient->packets_total = filesize ? (filesize / DCC_BLOCK_SIZE + 1) : 0;
	NewClient->packets_transfer = 0;
	time(&NewClient->lasttime);
	if (dcc_count == 0) dcc_count = 1;
	NewClient->dccnum = dcc_count++;
	return NewClient;
}

/*
 * Added by Chaos: Is used in edit.c for checking redirect.
 */
#ifdef __STDC__
extern int	dcc_active (char *user)
#else
extern int	dcc_active (user)
	char	*user;
#endif
{
	return (dcc_searchlist("chat", user, DCC_CHAT, 0, (char *) 0, 1)) ? 1 : 0;
}


#ifdef __STDC__
static	void dcc_add_deadclient(DCC_list *client)
#else
static	void dcc_add_deadclient(client)
	DCC_list *client;
#endif
{
	struct deadlist *new;

	new = (struct deadlist *) new_malloc(sizeof(struct deadlist));
	new->next = deadlist;
	new->it = client;
	deadlist = new;
}

/*
 * dcc_erase searches for the given entry in the dcc_list and
 * removes it
 */
#ifdef __STDC__
static void	dcc_erase(DCC_list *Element)
#else
static void dcc_erase(Element)
	DCC_list	*Element;
#endif
{
	DCC_list	**Client;
	dcc_count = 1;
	
	for (Client = &ClientList; *Client; Client = &(**Client).next)
	{
		if (*Client == Element)
		{
			*Client = Element->next;
			if (Element->write != -1)
				close(Element->write);
			if (Element->read != -1)
			{
				FD_CLR(Element->read, &readables);
				close(Element->read);
			}
			if (Element->file != -1)
				close(Element->file);

			new_free(&Element->description);
			new_free(&Element->user);
			new_free(&Element->buffer);
			new_free((char **)&Element);
			break;
		}
	}
	for (Client = &ClientList; *Client; Client = &(**Client).next)
		(*Client)->dccnum = dcc_count++;
}

static void dcc_really_erase _((void))
{
	struct deadlist *dies;

	for (;(dies = deadlist) != NULL;)
	{
		deadlist = deadlist->next;
		dcc_erase(dies->it);
	}
}

/*
 * Set the descriptor set to show all fds in Client connections to
 * be checked for data.
 */
#ifdef __STDC__
extern void	set_dcc_bits (fd_set *rd, fd_set *wd)
#else
extern void set_dcc_bits(rd, wd)
	fd_set	*rd, *wd;
#endif
{
	DCC_list	*Client;

	for (Client = ClientList; Client != NULL; Client = Client->next)
	{
		if (Client->read != -1)
			FD_SET(Client->read, rd);
#ifdef DCC_CNCT_PEND
		if (Client->write != -1 && (Client->flags & DCC_CNCT_PEND))
			FD_SET(Client->write, wd);
#endif
	}
}

#ifdef NON_BLOCKING_CONNECTS
static	void
dcc_got_connected(client)
	DCC_list *client;
{
	client->starttime = time(NULL);
	client->flags &= ~DCC_CNCT_PEND;
	set_blocking(client->read);
	if (client->read != client->write)
		set_blocking(client->write);
	if (client->flags & DCC_OFFER)
	{
		struct sockaddr_in	remaddr = {0};
		int	rl = sizeof(remaddr);

		client->flags &= ~DCC_OFFER;
		if (getpeername(client->read, (struct sockaddr *) &remaddr, &rl) == -1)
			say("DCC error: getpeername failed: %s - continuing with fingers crossed", strerror(errno));
		if ((client->flags & DCC_TYPES) != DCC_RAW)
			if (do_hook(DCC_CONNECT_LIST, "%s %s %s %d", client->user, 
				dcc_types[client->flags&DCC_TYPES],inet_ntoa(remaddr.sin_addr),
				ntohs(remaddr.sin_port))) 
			{
			say("DCC %s connection with %s[%s,%d] established",
				dcc_types[client->flags&DCC_TYPES], client->user,
				inet_ntoa(remaddr.sin_addr), ntohs(remaddr.sin_port));
			}
		message_from(NULL, LOG_CURRENT);
		if (((client->flags & DCC_REGETFILE) == DCC_REGETFILE) && client->flags)
		{
		        /* send a packet to the sender with transfer resume instructions */
			say("Telling uplink we want to start at %ld", client->transfer_orders.byteoffset);
        	        send(client->read, (const char *)&client->transfer_orders, sizeof(struct transfer_struct), 0);
		}
	}
}
#endif

/*
 * Check all DCCs for data, and if they have any, perform whatever
 * actions are required.
 */
#ifdef __STDC__
extern void	dcc_check (fd_set *Readables, fd_set *Writables)
#else
extern void dcc_check(Readables)
	fd_set	*Readables;
	fd_set	*Writables;
#endif
{
	DCC_list	**Client;
	struct	timeval	timeout;
	int	previous_server;
	int	lastlog_level;

	previous_server = from_server;
	from_server = (-1);
	timeout.tv_sec = timeout.tv_usec = 0;
	message_from(NULL, LOG_DCC);
	lastlog_level = set_lastlog_msg_level(LOG_DCC);
	for (Client = (&ClientList); *Client != NULL;)
	{
#ifdef NON_BLOCKING_CONNECTS
		if ((*Client)->flags & DCC_CNCT_PEND && (*Client)->write != -1 && FD_ISSET((*Client)->write, Writables))
		{
			/*
			 * If we were waiting for the connection to get
			 * established, we have some final setup to do.
			 */
			dcc_got_connected(*Client);
		}
#endif
		if ((*Client)->read != -1 && FD_ISSET((*Client)->read, Readables))
		{
#ifdef NON_BLOCKING_CONNECTS
			if ((*Client)->flags & DCC_CNCT_PEND)
			{
				dcc_got_connected(*Client);
				goto done;
			}
#endif
			switch((*Client)->flags&DCC_TYPES)
			{
			case DCC_CHAT:
				process_incoming_chat(*Client);
				break;
			case DCC_RAW_LISTEN:
				process_incoming_listen(*Client);
				break;
			case DCC_RAW:
				process_incoming_raw(*Client);
				break;
			case DCC_FILEOFFER:
			case DCC_RESENDOFFER:
				process_outgoing_file(*Client);
				break;
			case DCC_FILEREAD:
			case DCC_REGETFILE:
				process_incoming_file(*Client);
				break;
			}
		}
#ifdef NON_BLOCKING_CONNECTS
done:
#endif
		/*
		 * added for the case when a /dcc close comes before we
		 * reach the top of the loop.
		 */
		if (!*Client)
			break;

		if ((*Client)->flags & DCC_DELETE)
		{
			dcc_add_deadclient(*Client);
			Client = (&(**Client).next);
		}
		else
			Client = (&(**Client).next);
	}

	message_from(NULL, LOG_CRAP);
	(void) set_lastlog_msg_level(lastlog_level);
	dcc_really_erase();
	from_server = previous_server;
}

/*
 * Process a DCC command from the user.
 */
#ifdef __STDC__
extern void	process_dcc(char *args)
#else
extern void process_dcc(args)
	char	*args;
#endif
{
	char	*command;
	int	i;
	int	lastlog_level;

	if (!(command = next_arg(args, &args)))
		return;
	for (i = 0; dcc_commands[i].name != NULL; i++)
	{
		if (!my_stricmp(dcc_commands[i].name, command))
		{
			message_from(NULL, LOG_DCC);
			lastlog_level = set_lastlog_msg_level(LOG_DCC);
			dcc_commands[i].function(args);
			message_from(NULL, LOG_CURRENT);
			(void) set_lastlog_msg_level(lastlog_level);
			return;
		}
	}
	say("Unknown DCC command: %s", command);
}

#ifdef __STDC__
static	int	dcc_open (DCC_list *Client)
#else
static	int dcc_open(Client)
	DCC_list	*Client;
#endif
{
		char    *	user,
			*	Type;
	struct	in_addr 	myip;
#ifndef NON_BLOCKING_CONNECTS
	struct	sockaddr_in	remaddr;
		int		rl = sizeof(remaddr);
#endif
		int		old_server;

	user = Client->user;
	old_server = from_server;
	if (from_server == -1)
		from_server = get_window_server(0);
	myip.s_addr = server_list[from_server].local_addr.s_addr;
	Type = dcc_types[Client->flags & DCC_TYPES];
	if (Client->flags & DCC_OFFER)
	{
		message_from(NULL, LOG_DCC);
#ifdef DCC_CNCT_PEND
		Client->flags |= DCC_CNCT_PEND;
#endif
	/*BLAH!*/
		Client->remport = ntohs(Client->remport);
		if ((Client->write = connect_by_number(inet_ntoa(Client->remote), &Client->remport, SERVICE_CLIENT, PROTOCOL_TCP, 1)) < 0)
		{
			say("Unable to create connection: %s", errno ? sys_errlist[errno] : "Unknown Host");
			message_from((char *) 0, LOG_CURRENT);
			dcc_erase(Client);
			from_server = old_server;
			return 0;
		}
	/* BLAH! */
		Client->remport = htons(Client->remport);
		Client->read = Client->write;
		FD_SET(Client->read, &readables);
		Client->flags |= DCC_ACTIVE;
#ifndef NON_BLOCKING_CONNECTS
		Client->flags &= ~DCC_OFFER;
		(void) getpeername(Client->read, (struct sockaddr *) &remaddr, &rl);
		if (get_to_from(Type) != -1)
			dcc_active_count++;
		if ((Client->flags & DCC_TYPES) != DCC_RAW)
		{
                        if (do_hook(DCC_CONNECT_LIST,"%s %s %s %d",
                                user, Type, inet_ntoa(remaddr.sin_addr),
                                ntohs(remaddr.sin_port)))
			say("DCC %s connection with %s[%s %d] established",
				Type, user, inet_ntoa(remaddr.sin_addr),
				ntohs(remaddr.sin_port));
		}
		message_from(NULL, LOG_CURRENT);
#endif	
		Client->starttime = time(NULL);
		Client->bytes_read = Client->bytes_sent = 0L;
		from_server = old_server;
		return 1;
	}
	else
	{
		unsigned short portnum = 0;

#ifdef DCC_DCNT_PEND
		Client->flags |= DCC_WAIT|DCC_CNCT_PEND;
#else
		Client->flags |= DCC_WAIT;
#endif
		message_from(NULL, LOG_DCC);
		if ((Client->read = connect_by_number(NULL, &portnum, SERVICE_SERVER, PROTOCOL_TCP, 1)) < 0)
		{
			say("Unable to create connection: %s", errno ? sys_errlist[errno] : "Unknown Host");
			message_from(NULL, LOG_CURRENT);
			dcc_erase(Client);
			from_server = old_server;
			return 0;
		}
		if (get_to_from(Type) != -1)
			dcc_active_count++;
		if (Client->flags & DCC_TWOCLIENTS)
		{
			/* patch to NOT send pathname accross */
			char	*nopath;

			if (((Client->flags & DCC_FILEOFFER)== DCC_FILEOFFER) && (nopath = rindex(Client->description, '/')))
				nopath++;
			else
				nopath = Client->description;

			if (Client->filesize)
			{
				unsigned char foo[4];
				file_checksum(foo, Client->description);
				Client->cksum = m_sprintf("%x%x%x%x", foo[0], foo[1], foo[2], foo[3]);
				send_ctcp(ctcp_type[in_ctcp_flag], user, "DCC",
					 "%s %s %lu %u %d %s", Type, nopath,
					 (u_long) ntohl(myip.s_addr),
					 (u_short) portnum,
					 Client->filesize, Client->cksum);
			}
			else
				send_ctcp(ctcp_type[in_ctcp_flag], user, "DCC",
					 "%s %s %lu %u", Type, nopath,
					 (u_long) ntohl(myip.s_addr),
					 (u_short) portnum);
			message_from(NULL, LOG_DCC);
			if (!doing_multi && !dcc_quiet)
				say("Sent DCC %s request to %s", Type, user);
			message_from(NULL, LOG_CURRENT);
		}
		FD_SET(Client->read, &readables);
		Client->starttime = 0;
		from_server = old_server;
		return 2;
	}
}

#ifdef __STDC__
void	dcc_chat (char *args)
#else
void dcc_chat(args)
	char	*args;
#endif
{
	char	*user;
	DCC_list	*Client;

	if ((user = next_arg(args, &args)) == NULL)
	{
		say("You must supply a nickname for DCC CHAT");
		return;
	}
	Client = dcc_searchlist("chat", user, DCC_CHAT, 1, NULL, -1);
	if ((Client->flags & DCC_ACTIVE) || (Client->flags & DCC_WAIT))
	{
		say("A previous DCC CHAT to %s exists", user);
		return;
	}
	Client->flags |= DCC_TWOCLIENTS;
	dcc_open(Client);
}

#ifdef __STDC__
extern char	*dcc_raw_listen (int port)
#else
extern char	* dcc_raw_listen(port)
	int	port;
#endif
{
	DCC_list	*Client;
	char	*PortName;
	struct	sockaddr_in locaddr;
	char	*RetName = NULL;
	int	size;
	int	lastlog_level;

	lastlog_level = set_lastlog_msg_level(LOG_DCC);
	if (port && port < 1025)
	{
		say("Cannot bind to a privileged port");
		(void) set_lastlog_msg_level(lastlog_level);
		return NULL;
	}
	PortName = ltoa(port);
	Client = dcc_searchlist("raw_listen", PortName, DCC_RAW_LISTEN, 1, NULL, -1);
	if (Client->flags & DCC_WAIT)
	{
		say("A previous DCC RAW_LISTEN on %s exists", PortName);
		set_lastlog_msg_level(lastlog_level);
		return RetName;
	}
	bzero((char *) &locaddr, sizeof(locaddr));
	locaddr.sin_family = AF_INET;
	locaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	locaddr.sin_port = htons(port);
	if (0 > (Client->read = socket(AF_INET, SOCK_STREAM, 0)))
	{
		dcc_erase(Client);
		say("socket() failed: %s", sys_errlist[errno]);
		(void) set_lastlog_msg_level(lastlog_level);
		return RetName;
	}
	FD_SET(Client->read, &readables);
	set_socket_options(Client->read);
	if (bind(Client->read, (struct sockaddr *) &locaddr, sizeof(locaddr)) == -1)
	{
		dcc_erase(Client);
		say("Could not bind port: %s", sys_errlist[errno]);
		(void) set_lastlog_msg_level(lastlog_level);
		return RetName;
	}
	listen(Client->read, 4);
	size = sizeof(locaddr);
	Client->starttime = time(NULL);
	getsockname(Client->read, (struct sockaddr *) &locaddr, &size);
	Client->write = ntohs(locaddr.sin_port);
	Client->flags |= DCC_ACTIVE;
	malloc_strcpy(&Client->user, ltoa(Client->write));
	malloc_strcpy(&RetName, Client->user);
	(void) set_lastlog_msg_level(lastlog_level);
	return RetName;
}

#ifdef __STDC__
extern char	*dcc_raw_connect(char *host, u_short port)
#else
extern char	* dcc_raw_connect(host, port)
	char	*host;
	u_short	port;
#endif
{
	DCC_list	*Client;
	char	*PortName;
	struct	in_addr	address;
	struct	hostent	*hp;
	int	lastlog_level;

	lastlog_level = set_lastlog_msg_level(LOG_DCC);
	if ((address.s_addr = inet_addr(host)) == -1)
	{
		hp = gethostbyname(host);
		if (!hp)
		{
			say("Unknown host: %s", host);
			(void) set_lastlog_msg_level(lastlog_level);
			return m_strdup(empty_string);
		}
		bcopy(hp->h_addr, &address, sizeof(address));
	}
	Client = dcc_searchlist(host, ltoa(port), DCC_RAW, 1, NULL, -1);
	if (Client->flags & DCC_ACTIVE)
	{
		say("A previous DCC RAW to %s on %s exists", host, ltoa(port));
		(void) set_lastlog_msg_level(lastlog_level);
		return m_strdup(empty_string);
	}
	/* Sorry. The missing 'htons' call here broke $connect() */
	Client->remport = htons(port);
	bcopy((char *) &address, (char *) &Client->remote, sizeof(address));
	Client->flags = DCC_OFFER | DCC_RAW;
	if (!dcc_open(Client))
		return m_strdup(empty_string);
	PortName = ltoa(Client->read);
	Client->user = m_strdup(PortName);
	if (do_hook(DCC_RAW_LIST, "%s %s E %d", PortName, host, port))
            if (do_hook(DCC_CONNECT_LIST,"%s RAW %s %d",PortName, host, port))
		say("DCC RAW connection to %s on %s via %d established", host, PortName, port);

	(void) set_lastlog_msg_level(lastlog_level);
	return m_strdup(PortName);
}


void	multisend(usern, filen)
char	*usern, *filen;
{
	char *newbuf = NULL;
	char *filebuf = NULL;
	char *path = NULL;
	char *thefile;
	DIR *dp;
	char *expand;
	struct dirent *dir;
	struct stat stbuf;

	doing_multi = 1;
	malloc_strcpy(&path, filen);

	if ((thefile = strrchr(path, '/')) != NULL)
		*thefile++ = '\0';
	else {
		malloc_strcpy(&path, "~");
		thefile = filen;
	}
	expand = expand_twiddle(path);
	
	dp = opendir(expand);
	if ((dp = opendir(expand)) == NULL)
	{
		say("Cannot access directory %s!", expand);
		new_free(&expand);
		new_free(&path);
		doing_multi = 0;
		return;
	}
	while (1)
	{
		dir = readdir(dp);
		if (!dir) {
			new_free(&expand);
			new_free(&path);
			doing_multi = 0;
			return;
		}
		if (dir->d_ino == 0)
			continue;
		if (!wild_match(thefile, dir->d_name)||!wild_match(dir->d_name, thefile))
			continue;
		malloc_sprintf(&filebuf, "%s/%s", expand, dir->d_name);
		stat(filebuf, &stbuf);
		if (S_ISDIR(stbuf.st_mode))
			continue;

		malloc_sprintf(&newbuf, "%s %s/%s", usern, expand, dir->d_name);
		say("Sent DCC SEND request to %s for file %s", usern, dir->d_name);
		dcc_filesend(newbuf);
	}
	new_free(&newbuf);
	new_free(&filebuf);
	new_free(&path);
	doing_multi = 0;
}

void
dcc_resend(args)
	char	*args;
{
	char	*user;
	char	*filename,
		*fullname;
	DCC_list *Client;
	char	*FileBuf = NULL;
	struct	stat	stat_buf;

	if (!(user = next_arg(args, &args)) ||
	    !(filename = next_arg(args, &args)))
	{
		say("You must supply a nickname and filename for DCC RESEND");
		return;
	}
	if (*filename == '/')
	{
		malloc_strcpy(&FileBuf, filename);
	}
	else if (*filename == '~')
	{
		if (!(fullname = expand_twiddle(filename)))
		{
			yell("Unable to expand %s", filename);
			return;
		}
		malloc_strcpy(&FileBuf, fullname);
		new_free(&fullname);
	}
	else
	{
		char current_dir[BIG_BUFFER_SIZE+1];
		getcwd(current_dir, sizeof(current_dir) - strlen(filename) - 4);
		malloc_sprintf(&FileBuf, "%s/%s", current_dir, filename); 
	}
	if (0 != access(FileBuf, R_OK))
	{
		yell("Cannot access %s", FileBuf);
		new_free(&FileBuf);
		return;
	}
	stat_file(FileBuf, &stat_buf);
	if (stat_buf.st_mode & S_IFDIR)
	{
		yell("Cannot send a directory");
		new_free(&FileBuf);
		return;
	}
	if (!my_strnicmp(FileBuf, "/etc/", 5))
	{
		yell("Resend request rejected");
		new_free(&FileBuf);
		return;
	}

	filesize = stat_buf.st_size;
	Client = dcc_searchlist(FileBuf, user, DCC_RESENDOFFER, 1, filename,-1);
	filesize = 0;
	if ((Client->flags & DCC_ACTIVE) || (Client->flags & DCC_WAIT))
	{
		say("A previous DCC RESEND:%s to %s exists", FileBuf, user);
		new_free(&FileBuf);
		return;
	}
	Client->flags |= DCC_TWOCLIENTS;
	dcc_open(Client);
	dcc_show_active(NULL);
	new_free(&FileBuf);
}

#ifdef __STDC__
void		dcc_filesend (char *args)
#else
void dcc_filesend(args)
	char	*args;
#endif
{
	char	*user;
	char	*filename,
		*fullname;
	DCC_list *Client;
	char	FileBuf[DCC_BLOCK_SIZE+1];
	struct	stat	stat_buf;

	if (!(user = next_arg(args, &args)) || !(filename = next_arg(args, &args)))
	{
		say("You must supply a nickname and filename for DCC SEND");
		return;
	}
	if (*filename == '/')
		strcpy(FileBuf, filename);

	else if (*filename == '~')
	{
		if (0 == (fullname = expand_twiddle(filename)))
		{
			yell("Unable to expand %s", filename);
			return;
		}
		strcpy(FileBuf, fullname);
		new_free(&fullname);
	}
	else
	{
		getcwd(FileBuf, sizeof(FileBuf));
		strcat(FileBuf, "/");
		strcat(FileBuf, filename);
	}

	if (0 != access(FileBuf, R_OK))
	{
		yell("Cannot access %s", FileBuf);
		return;
	}
	stat_file(FileBuf, &stat_buf);
	if (stat_buf.st_mode & S_IFDIR)
	{
		yell("Cannot send a directory");
		return;
	}
	if (scanstr(FileBuf, "/etc/"))
	{
		yell("Send request rejected");
		return;
	}
	if (strlen(FileBuf) >= 7 &&
	    0 == strcmp(FileBuf + strlen(FileBuf) - 7, "/passwd"))
	{
		yell("Send request rejected");
		return;
	}
	filesize = stat_buf.st_size;
	Client = dcc_searchlist(FileBuf, user, DCC_FILEOFFER, 1, filename, -1);
	filesize = 0;
	if ((Client->flags & DCC_ACTIVE) || (Client->flags & DCC_WAIT))
	{
		say("A previous DCC SEND:%s to %s exists", FileBuf, user);
		return;
	}
	Client->flags |= DCC_TWOCLIENTS;
	dcc_open(Client);
}

void	multiget(usern, filen)
char	*usern, *filen;
{
	DCC_list	*dccList;
	char		*newbuf = NULL;
	char 		*expand = NULL;	
	doing_multi = 1;
	if (strchr(filen, ' '))
		*strchr(filen, ' ') = '\0';
	for (dccList = ClientList; dccList ; dccList = dccList->next)
        {
        	if((wild_match(dccList->description, filen) || wild_match(filen, dccList->description))&& 
			!my_stricmp(usern, dccList->user) && 
			(dccList->flags & DCC_TYPES) == DCC_FILEREAD)
		{
			if (get_string_var(DCC_DLDIR_VAR)) {
				expand = expand_twiddle(get_string_var(DCC_DLDIR_VAR));
				malloc_sprintf(&newbuf, "%s %s/%s", usern, expand, dccList->description);
			} else
				malloc_sprintf(&newbuf, "%s %s", usern, dccList->description);
			say("Attempting DCC GET of file %s", dccList->description);
			dcc_getfile(newbuf);
			new_free(&expand);
			new_free(&newbuf);
		}
	}
	doing_multi = 0;
}
 

#ifdef __STDC__
static	void		dcc_getfile (char *args)
#else
static void dcc_getfile(args)
	char	*args;
#endif
{
	char	*user;
	char	*filename;
	char	*tmp = NULL;
	DCC_list	*Client;
	char	*fullname = NULL;

	if (0 == (user = next_arg(args, &args)))
	{
		say("You must supply a nickname for DCC GET");
		return;
	}
	filename = next_arg(args, &args);

	if (0 == (Client = dcc_searchlist(filename, user, DCC_FILEREAD, 0, (char *) 0, 0)))
	{
		if (filename)
			say("No file (%s) offered in SEND mode by %s", filename, user);
		else
			say("No file offered in SEND mode by %s", user);
		return;
	}
	if ((Client->flags & DCC_ACTIVE) || (Client->flags & DCC_WAIT))
	{
		say("A previous DCC GET:%s to %s exists", filename?filename:"<any>", user);
		return;
	}
	if (0 == (Client->flags & DCC_OFFER))
	{
		say("I'm a teapot!");
		dcc_erase(Client);
		return;
	}
	Client->flags |= DCC_TWOCLIENTS;
	Client->bytes_sent = Client->bytes_read = 0L;

	if (!dcc_open(Client))
		return;
	if (get_string_var(DCC_DLDIR_VAR))
		malloc_sprintf(&tmp, "%s/%s", get_string_var(DCC_DLDIR_VAR), Client->description);
	else
		malloc_sprintf(&tmp, "%s", Client->description);
	if (0 == (fullname = expand_twiddle(tmp)))
		/* left over phone bug */
		malloc_strcpy(&fullname, tmp);


	if (-1 == (Client->file = open(fullname, O_WRONLY | O_TRUNC | O_CREAT, 0644)))
	{
		say("Unable to open %s: %s", Client->description, errno ? sys_errlist[errno] : "<No Error>");
		FD_CLR(Client->read, &readables);
		close(Client->read);
		dcc_erase(Client);
	}
	new_free(&fullname);
	new_free(&tmp);
}


void dcc_regetfile(args)
	char	*args;
{
	char	*user;
	char	*tmp = NULL;
	char	*filename;
	DCC_list	*Client;
	char	*fullname = NULL;
	struct stat buf;

	if (0 == (user = next_arg(args, &args)))
	{
		say("You must supply a nickname for DCC REGET");
		return;
	}
	filename = next_arg(args, &args);

	if (0 == (Client = dcc_searchlist(filename, user, DCC_REGETFILE, 0, NULL, 0)))
	{
		if (filename)
			say("No file (%s) offered in RESEND mode by %s",
					filename, user);
		else
			say("No file offered in RESEND mode by %s", user);
		return;
	}
	if ((Client->flags & DCC_ACTIVE) || (Client->flags & DCC_WAIT))
	{
		say("A previous DCC REGET:%s to %s exists", filename, user);
		return;
	}
	if (0 == (Client->flags & DCC_OFFER))
	{
		say("I'm a teapot!");
		dcc_erase(Client);
		return;
	}
	Client->flags |= DCC_TWOCLIENTS;
	Client->bytes_sent = Client->bytes_read = 0L;

	if (get_string_var(DCC_DLDIR_VAR))
		malloc_sprintf(&tmp, "%s/%s", get_string_var(DCC_DLDIR_VAR), Client->description);
	else
		malloc_sprintf(&tmp, "%s", Client->description);
	if (0 == (fullname = expand_twiddle(tmp)))
		malloc_strcpy(&fullname, tmp);

	if (!dcc_open(Client))
		return;
	if (-1 == (Client->file = open(fullname, O_WRONLY | O_CREAT, 0644)))
	{
		say("Unable to open %s: %s", Client->description,
				errno ? strerror(errno) : "<No Error>");
		close(Client->read);
		dcc_erase(Client);
	}

	/* seek to the end of the file about to be resumed */
	lseek(Client->file, 0, SEEK_END);

	/* get the size of our file to be resumed */
	fstat(Client->file, &buf);

	Client->transfer_orders.packet_id = DCC_PACKETID;
	Client->transfer_orders.byteoffset = buf.st_size;
	Client->transfer_orders.byteorder = byteordertest();
#ifndef NON_BLOCKING_CONNECTS
	/* send a packet to the sender with transfer resume instructions */
	send(Client->read, (const char *)&(Client->transfer_orders), sizeof(struct transfer_struct), 0);
#endif
	new_free(&fullname);
	new_free(&tmp);
}

#ifdef __STDC__
extern void	register_dcc_offer (char *user, char *type, char *description, char *address, char *port, char *size, char *extra)
#else
extern void register_dcc_offer(user, type, description, address, port, size, extra)
	char	*user;
	char	*type;
	char	*description;
	char	*address;
	char	*port;
	char	*size;
	char	*extra;
#endif
{
	DCC_list *	Client;
	int		CType;
	char *		c;
	u_long		TempLong;
	unsigned	TempInt;
	int		do_auto = 0;	/* used in dcc chat collisions */
	int		do_autog = 1;
	long		packets = 0;
		
	if ((c = rindex(description, '/')))
		description = c + 1;
	if ('.' == *description)
		*description = '_';

	if (size && *size)
	{
		filesize = atoi(size);
		packets = filesize / DCC_BLOCK_SIZE + 1;
	}
	else
		packets = filesize = 0;

	if (!my_stricmp(type, "CHAT"))
		CType = DCC_CHAT;
	else if (!my_stricmp(type, "SEND"))
		CType = DCC_FILEREAD;
/* added by Barnabus */
	else if (!my_stricmp(type, "RESEND"))
		CType = DCC_REGETFILE;
/* end add */
        else
        {
                say("Unknown DCC %s (%s) received from %s", type, description, user);
                return;
        }

	Client = dcc_searchlist(description, user, CType, 1, (char *) 0, -1);
	filesize = 0;

	if (extra && *extra)
		Client->cksum = m_strdup(extra);

	if (Client->flags & DCC_WAIT)
	{
		FD_CLR(Client->read, &readables);
		close(Client->read);
		dcc_erase(Client);
		if (DCC_CHAT == CType)
		{
			Client = dcc_searchlist(description, user, CType, 1, (char *) 0, -1);
			do_auto = 1;
		}
		else
		{
			say("DCC %s collision for %s:%s", type, user, description);
			send_ctcp_reply(user, "DCC", "DCC %s collision occured while connecting to %s (%s)", type, nickname, description);
			return;
		}
	}
	if (Client->flags & DCC_ACTIVE)
	{
		say("Received DCC %s request from %s while previous session still active", type, user);
		return;
	}
	Client->flags |= DCC_OFFER;

	TempLong = strtoul(address, NULL, 10);
	Client->remote.s_addr = htonl(TempLong);
	TempInt = (unsigned) strtoul(port, NULL, 10);
	Client->remport = htons(TempInt);

	if (TempInt < 1024)
	{
		say("DCC %s (%s) request from %s rejected", type, description, user);
		dcc_erase(Client);
		return;
	}

#ifdef HACKED_DCC_WARNING
	/* This right here compares the hostname from the userhost stamped
	 * on the incoming privmsg to the address stamped in the handshake.
	 * We do not automatically reject any discrepencies, but warn the
	 * user instead to be cautious.
	 */
	{
		char tmpbuf[128];
		char *fromhost;
		unsigned long compare, compare2;
		struct hostent *hostent_fromhost;

		strcpy(tmpbuf, FromUserHost);
		fromhost = index(tmpbuf, '@') + 1;
		alarm(1);	/* dont block too long... */
		hostent_fromhost = gethostbyname(fromhost);
		alarm(0);
		if (!hostent_fromhost)
		{
			yell("Incoming handshake has an address [%s] that could not be figured out!", fromhost);
			yell("Please use caution in deciding whether to accept it or not");
		}
		else
		{
			compare = *((unsigned long *)hostent_fromhost->h_addr_list[0]);
			compare2 = inet_addr(fromhost);
			if ((compare != Client->remote.s_addr) &&
				(compare2 != Client->remote.s_addr))
			{
				say("WARNING: Fake dcc handshake detected! [%x]",Client->remote.s_addr);
				say("Unless you know where this dcc request is coming from");
				say("It is recommended you ignore it!");
			}
		}
	}
#endif

	if ((u_long) 0 == TempLong || 0 == Client->remport)
	{
		yell("DCC handshake from %s ignored becuase it had an null port or address", user);
		dcc_erase(Client);
		return;
	}
	if (((Client->flags & DCC_TYPES) == DCC_FILEREAD) && ((Client->flags & DCC_TYPES) != DCC_REGETFILE))
	{
		struct stat statit;
		char *tmp = NULL, *p;
		malloc_sprintf(&tmp, "%s/%s", get_string_var(DCC_DLDIR_VAR), Client->description);
		p = expand_twiddle(tmp);
		if ( !dcc_overwrite_var && get_int_var(DCC_AUTOGET_VAR) && ((do_autog = stat(p, &statit)) == 0) )
			/* the file exists. warning is generated */
			say("Warning: File %s exists: use /DCC RENAME if you dont want to overwrite existing file", p);
		new_free(&tmp); new_free(&p);
	}
  	if (do_auto)
  	{
                if (do_hook(DCC_CONNECT_LIST,"%s CHAT",user))
			say("DCC CHAT already requested by %s, connecting...", user);
  		dcc_chat(user);
  	}
        else if (do_hook(DCC_REQUEST_LIST,"%s %s %s %d",user, type, description, Client->filesize))
	{
/* Thanks, Tychy! (lherron@imageek.york.cuny.edu) */
		if (Client->filesize)
			say("DCC %s (%s %d) request received from %s!%s [%s:%d]",
			    type,description, Client->filesize, user, FromUserHost,
			    inet_ntoa(Client->remote),Client->remport);
		else
		        say("DCC %s (%s) request received from %s!%s [%s:%d]", 
                             type, description, user, FromUserHost,
                             inet_ntoa(Client->remote),Client->remport);
	}
	if ((get_int_var(DCC_AUTOGET_VAR)) && 
	    (Client->filesize/1024) < get_int_var(MAX_AUTOGET_SIZE_VAR) 
	    && (!my_stricmp(type, "SEND") || !my_stricmp(type, "RESEND")))
	{
		char *thebuf;
		if (!Client->filesize)
		{
			bitchsay("Caution Filesize is equal to zero!! No Autoget");
			return;
		}
		thebuf = (char *) new_malloc(strlen(user)+strlen(description)+4);
		sprintf(thebuf, "%s %s", user, description);
		if ((CType == DCC_REGETFILE) || do_autog)
			say("Auto-%sting file %s from %s", (CType == DCC_REGETFILE) ? "reget":"get", description, user);
		if (CType == DCC_REGETFILE)
			dcc_regetfile(thebuf);
		else {
			if (do_autog) 
				dcc_getfile(thebuf);
		}
		new_free(&thebuf);
	}
	return;
}

#ifdef __STDC__
static	void		process_incoming_chat (DCC_list *Client)
#else
static void process_incoming_chat(Client)
	DCC_list	*Client;
#endif
{
	struct	sockaddr_in	remaddr;
	int	sra;
	char	tmp[DCC_BLOCK_SIZE+1];
	char	*s, *bufptr;
	long	bytesread;
	int	old_timeout;
	int	len = 0;
	
	if (Client->flags & DCC_WAIT)
	{
		sra = sizeof(struct sockaddr_in);
		Client->write = accept(Client->read, (struct sockaddr *) &remaddr, &sra);
#ifdef  ESIX
		mark_socket(Client->write);
#endif
		FD_CLR(Client->read, &readables);
		close(Client->read);
		Client->read = Client->write;
		FD_SET(Client->read, &readables);
		Client->flags &= ~DCC_WAIT;
		Client->flags |= DCC_ACTIVE;
                if (do_hook(DCC_CONNECT_LIST, "%s CHAT %s %d", Client->user,
                         inet_ntoa(remaddr.sin_addr)), ntohs(remaddr.sin_port))
		say("DCC chat connection to %s[%s,%d] established", Client->user,
			inet_ntoa(remaddr.sin_addr), ntohs(remaddr.sin_port));
		Client->starttime = time(NULL);
		return;
	}
	old_timeout = dgets_timeout(1);
        s = Client->buffer;
        bufptr = tmp;
        if (s && *s)
        {
                len = strlen(s);
		if (len > BIG_BUFFER_SIZE - 1)
		{
			bitchsay("dcc buffer overrun on chat freeing data");
			new_free(&(Client->buffer));
		} 
		else
		{ 
	                strncpy(tmp, s, len);
			bufptr += len;
		}
        }
	bytesread = dgets(bufptr, DCC_BLOCK_SIZE - len, Client->read, NULL);
	(void) dgets_timeout(old_timeout);
	switch (bytesread)
	{
	case -1:
                add_to_dcc_buffer(Client, tmp);
                return;
	case 0:
	{
		char *real_tmp = ((dgets_errno == -1) ? "Remote End Closed Connection" : strerror(dgets_errno));
                if (do_hook(DCC_LOST_LIST, "%s CHAT %s", Client->user, real_tmp))
        		say("DCC CHAT connection to %s lost", Client->user, real_tmp);
		FD_CLR(Client->read, &readables);
		close(Client->read);
		Client->read = Client->write = -1;
		Client->flags |= DCC_DELETE;
		return;
	}
	default:
	{
		char userhost[80];
		char equal_nickname[80];

		new_free(&Client->buffer);
		tmp[strlen(tmp) - 1]='\0';
		Client->bytes_read += bytesread;
		message_from(Client->user, LOG_DCC);

		if (!strncmp(tmp, "CTCP_MESSAGE ", strlen("CTCP_MESSAGE ")))
		{
			strcpy(userhost, "Unknown@");
			strcat(userhost, inet_ntoa(remaddr.sin_addr));
			FromUserHost = userhost;
			strcpy(equal_nickname, "=");
			strcat(equal_nickname, Client->user);
			do_ctcp(equal_nickname, get_server_nickname(-1), tmp + strlen("CTCP_MESSAGE "));
		}
		else if (!strncmp(tmp, "CTCP_REPLY ", strlen("CTCP_REPLY ")))
		{
			strcpy(userhost, "Unknown@");
			strcat(userhost, inet_ntoa(remaddr.sin_addr));
			FromUserHost = userhost;
			strcpy(equal_nickname, "=");
			strcat(equal_nickname, Client->user);
			do_notice_ctcp(equal_nickname, get_server_nickname(-1), tmp + strlen("CTCP_REPLY "));
		}
		else if (do_hook(DCC_CHAT_LIST, "%s %s", Client->user, tmp))
		{
			if (away_set)
			{
				time_t	t;
				char	*msg = NULL;

				t = time(NULL);
				msg = (char *) new_malloc(strlen(tmp) + 20);
				sprintf(msg, "%s <%.16s>", tmp, ctime(&t));
				strcpy(tmp, msg);
				new_free(&msg);
			}
			put_it("=%s= %s", Client->user, tmp);
		}
		message_from(NULL, LOG_CURRENT);
		return;
	}
	}
}

#ifdef __STDC__
static	void		process_incoming_listen (DCC_list *Client)
#else
static void process_incoming_listen(Client)
	DCC_list	*Client;
#endif
{
	struct	sockaddr_in	remaddr;
	int	sra;
	char	FdName[10];
	DCC_list	*NewClient;
	int	new_socket;
	struct	hostent	*hp;
#if defined(__linux__) || defined(__sgi)
	const char	*Name;
#else
	char	*Name;
#endif

	sra = sizeof(struct sockaddr_in);
	new_socket = accept(Client->read, (struct sockaddr *) &remaddr, &sra);
	if (0 != (hp = gethostbyaddr((char *)&remaddr.sin_addr,
	    sizeof(remaddr.sin_addr), remaddr.sin_family)))
		Name = hp->h_name;
	else
		Name = inet_ntoa(remaddr.sin_addr);
#ifdef  ESIX
	mark_socket(new_socket);
#endif
	sprintf(FdName, "%d", new_socket);
	NewClient = dcc_searchlist((char *)Name, FdName, DCC_RAW, 1, NULL, 0);
	NewClient->starttime = time(NULL);
	NewClient->read = NewClient->write = new_socket;
	FD_SET(NewClient->read, &readables);
	NewClient->remote = remaddr.sin_addr;
	NewClient->remport = remaddr.sin_port;
	NewClient->flags |= DCC_ACTIVE;
	NewClient->bytes_read = NewClient->bytes_sent = 0L;
	if (do_hook(DCC_RAW_LIST, "%s %s N %d", NewClient->user,
						NewClient->description,
						Client->write))
        if (do_hook(DCC_CONNECT_LIST,"%s RAW %s %d", NewClient->user,
                                                     NewClient->description,
                                                     Client->write))
		say ("DCC RAW connection to %s on %s via %d established",
					NewClient->description,
					NewClient->user,
					Client->write);
}

#ifdef __STDC__
static	void		process_incoming_raw (DCC_list *Client)
#else
static void process_incoming_raw(Client)
	DCC_list	*Client;
#endif
{
	char	tmp[DCC_BLOCK_SIZE+1];
	char 	*s, *bufptr;
	long	bytesread;
	int     old_timeout;

        s = Client->buffer;
        bufptr = tmp;
        if (s && *s)
        {
                int     len = strlen(s);
 
                strncpy(tmp, s, len);
                bufptr += len;
        }
        old_timeout = dgets_timeout(1);
	switch(bytesread = dgets(bufptr, DCC_BLOCK_SIZE, Client->read, (char *)0))
	{
	case -1:
	{
                add_to_dcc_buffer(Client, tmp);
                return;
	}
	case 0:
	{
		if (do_hook(DCC_RAW_LIST, "%s %s C",
				Client->user, Client->description))
       		if (do_hook(DCC_LOST_LIST,"%s RAW %s", 
				Client->user, Client->description))
			say("DCC RAW connection to %s on %s lost",
				Client->user, Client->description);
		FD_CLR(Client->read, &readables);
		close(Client->read);
		Client->read = Client->write = -1;
		Client->flags |= DCC_DELETE;
		(void) dgets_timeout(old_timeout);
		return;
	}
	default:
	{
		new_free(&Client->buffer);
		tmp[strlen(tmp) - 1] = '\0';
		Client->bytes_read += bytesread;
		if (do_hook(DCC_RAW_LIST, "%s %s D %s",
				Client->user, Client->description, tmp))
			say("Raw data on %s from %s: %s",
				Client->user, Client->description, tmp);
		(void) dgets_timeout(old_timeout);
		return;
	}
	}
}


/* Now that ive mucked this function up, i should go back and fix it. */
#ifdef __STDC__
static void	process_outgoing_file (DCC_list *Client)
#else
static void process_outgoing_file(Client)
	DCC_list	*Client;
#endif
{
	struct	sockaddr_in	remaddr;
	int			sra;
	char			tmp[DCC_BLOCK_SIZE+1];
	u_32int_t		bytesrecvd;
	int			bytesread;
	char			*Type;
	unsigned char 		packet[BIG_BUFFER_SIZE+1];
	struct  transfer_struct *received;
	int			BlockSize;
	
	Type = dcc_types[Client->flags & DCC_TYPES];
	if (Client->flags & DCC_WAIT)
	{
		sra = sizeof(struct sockaddr_in);
		Client->write = accept(Client->read, (struct sockaddr *) &remaddr, &sra);
#ifdef  ESIX
		mark_socket(Client->write);
#endif
		if ((Client->flags & DCC_RESENDOFFER) == DCC_RESENDOFFER)
		{
			recv(Client->write, packet, sizeof(struct transfer_struct), 0);
			received = (struct transfer_struct *)packet;
			if (byteordertest() != received->byteorder)
			{
				/* the packet sender orders bytes differently than us,
				 * reverse what they sent to get the right value 
				 */
				Client->transfer_orders.packet_id = 
					((received->packet_id & 0x00ff) << 8) | 
					((received->packet_id & 0xff00) >> 8);

				 Client->transfer_orders.byteoffset = 
					((received->byteoffset & 0xff000000) >> 24) |
					((received->byteoffset & 0x00ff0000) >> 8) |
					((received->byteoffset & 0x0000ff00) << 8) |
					((received->byteoffset & 0x000000ff) << 24);
			}
			else
			{
				memcpy(&Client->transfer_orders,packet,sizeof(struct transfer_struct));
			}

			if (Client->transfer_orders.packet_id != DCC_PACKETID)
			{
				say("DCC REGET packet is invalid!");
			}
		}
   /* end add */
		FD_CLR(Client->read, &readables);
		close(Client->read);
		Client->read = Client->write;
		FD_SET(Client->read, &readables);
		Client->flags &= ~DCC_WAIT;
		Client->flags |= DCC_ACTIVE;
		Client->bytes_sent = 0L;
		Client->starttime = time(NULL);
		if ((Client->file = open(Client->description, O_RDONLY)) == -1)
		{
			say("Unable to open %s: %s\n", Client->description,
				errno ? sys_errlist[errno] : "Unknown Host");
			FD_CLR(Client->read, &readables);
			close(Client->read);
			Client->read = Client->write = (-1);
			Client->flags |= DCC_DELETE;
			return;
		}
		if ((Client->flags & DCC_RESENDOFFER) == DCC_RESENDOFFER)		
		{
			lseek(Client->file, Client->transfer_orders.byteoffset, SEEK_SET);
			errno = 0;
			if (do_hook(DCC_CONNECT_LIST,"%s RESEND %s %d",Client->user,
				inet_ntoa(remaddr.sin_addr), ntohs(remaddr.sin_port)))
				if (dcc_quiet == 0)
					say("DCC SEND connection to %s[%s,%d] at %ld bytes established", Client->user,
						inet_ntoa(remaddr.sin_addr), ntohs(remaddr.sin_port), Client->transfer_orders.byteoffset);
		}
		else
		{
			if (do_hook(DCC_CONNECT_LIST,"%s SEND %s %d",Client->user,
				inet_ntoa(remaddr.sin_addr), ntohs(remaddr.sin_port)))
				if (dcc_quiet == 0)
					say("DCC SEND connection to %s[%s,%d] established", Client->user,
						inet_ntoa(remaddr.sin_addr), ntohs(remaddr.sin_port));
		}
	}
	else 
	{ 
		if (recv(Client->read, (char *)&bytesrecvd, sizeof(u_32int_t), 0) < sizeof(u_32int_t))
		{
                        if (do_hook(DCC_LOST_LIST,"%s SEND %s CONNECTION LOST",
                                Client->user, Client->description))
	       		say("DCC SEND:%s connection to %s lost",
				Client->description, Client->user);
			FD_CLR(Client->read, &readables);
			close(Client->read);
			if (get_to_from(Type) != -1 && dcc_active_count)
				dcc_active_count--;
			DCC_done = 1;
			Client->read = Client->write = (-1);
			Client->flags |= DCC_DELETE;
			close(Client->file);
			dcc_update_stats(Client);
			status_update(1);
			return;
		}
		else if (ntohl(bytesrecvd) != Client->bytes_sent)
			return;
	}

	BlockSize = get_int_var(DCC_BLOCK_SIZE_VAR);
	if (BlockSize > DCC_BLOCK_SIZE)
		BlockSize = DCC_BLOCK_SIZE;
	else if (BlockSize < 16)
		BlockSize = 16;
	if ((bytesread = read(Client->file, tmp, DCC_BLOCK_SIZE)) != 0)
	{
		send(Client->write, tmp, bytesread, 0);
		Client->packets_transfer++;
		Client->bytes_sent += bytesread;
		if (Client->filesize)
			update_transfer_buffer("[to %s: %d of %d %d%%]", Client->user, Client->packets_transfer, Client->packets_total, Client->bytes_sent * 100 / (Client->filesize-Client->transfer_orders.byteoffset));
		else
			update_transfer_buffer("[to %s: %d of %d %dK]", Client->user, Client->packets_transfer, Client->packets_total, Client->bytes_sent / 1024);
		DCC_done = 0;
		status_update(1);
	}
	else
		DCC_close_filesend(Client);
}

#ifdef __STDC__
static	void		process_incoming_file (DCC_list *Client)
#else
static void process_incoming_file(Client)
	DCC_list	*Client;
#endif
{
	char		tmp[DCC_BLOCK_SIZE+1];
	u_32int_t	bytestemp;
	int		bytesread;
	char		*Type;
	
	Type = dcc_types[Client->flags & DCC_TYPES];
	if ((bytesread = recv(Client->read, tmp, DCC_BLOCK_SIZE, 0)) <= 0)
	{
		unsigned char foo[4];
		char tmpbuf[64];

		if (Client->cksum)
		{
			file_checksum(foo, Client->description);
			sprintf(tmpbuf, "%x%x%x%x", foo[0], foo[1], foo[2], foo[3]);
			if (strcmp(Client->cksum, tmpbuf))
				yell("Incoming file's checksum does not match [%s] [%s]", Client->cksum, tmpbuf);
			else
				yell("Incoming file has been recieved with no errors.");
		}
		DCC_close_filesend(Client);
	}
	else
	{
		write(Client->file, tmp, bytesread);
		Client->bytes_read += bytesread;
		bytestemp = htonl(Client->bytes_read);
		send(Client->write, (char *)&bytestemp, sizeof(u_32int_t), 0);
		Client->packets_transfer++;
/* TAKE THIS OUT IF IT CAUSES PROBLEMS */
		if ((Client->filesize) && (Client->bytes_read > Client->filesize))
		{
			bitchsay("DCC GET WARNING: incoming file is larger then the handshake said");
			bitchsay("DCC GET: Closing connection");
			FD_CLR(Client->read, &readables);
			close(Client->read);
			close(Client->file);
			Client->read = Client->write = (-1);
			Client->flags |= DCC_DELETE;
			dcc_update_stats(Client);
			return;
		}
		if (Client->filesize) 
		{
			if (Client->filesize)
				update_transfer_buffer("[%s %d of %d %d%%]", Client->user, Client->packets_transfer, Client->packets_total, (int)(Client->bytes_read * 100 / Client->filesize));
			else
				update_transfer_buffer("[%s %d of %d %dK]", Client->user, Client->packets_transfer, Client->packets_total, Client->filesize / 1024);
		}
	}
}

/* flag == 1 means show it.  flag == 0 used by redirect (and /ctcp) */

#ifdef __STDC__
extern void	dcc_message_transmit (char *user, char *text, int type, int flag)
#else
extern void dcc_message_transmit(user, text, type, flag)
	char	*user;
	char	*text;
	int	type,
		flag;
#endif
{
	DCC_list	*Client;
	char	tmp[DCC_BLOCK_SIZE+1];
	int	lastlog_level;
	char	thing = '\0';
	int	list = 0;
	char	*host = NULL;

	switch(type)
	{
		case DCC_CHAT:
			host = "chat";
			thing = '=';
			list = SEND_DCC_CHAT_LIST;
			break;
		case DCC_RAW:
			host = next_arg(text, &text);
			if (!host)
			{
				say("No host specified for DCC RAW");
				return;
			}
			break;
	}
	if (!(Client = dcc_searchlist(host, user, type, 0, NULL, 1)) || !(Client->flags&DCC_ACTIVE))
	{
		say("No active DCC %s:%s connection for %s",
			dcc_types[type], host?host:"<any>", user);
		return;
	}
	lastlog_level = set_lastlog_msg_level(LOG_DCC);
	message_from(Client->user, LOG_DCC);
	strmcpy(tmp, text, DCC_BLOCK_SIZE);
	strmcat(tmp, "\n", DCC_BLOCK_SIZE); 
	send(Client->write, tmp, strlen(tmp), 0);
	Client->bytes_sent += strlen(tmp);
	if (flag)
	{
		if (type != DCC_RAW)
			if (do_hook(list, "%s %s", Client->user, text))
				put_it("=> %c%s%c %s", thing, Client->user, thing, text);
	}
	set_lastlog_msg_level(lastlog_level);
	message_from(NULL, LOG_CURRENT);
	return;
}

#ifdef __STDC__
extern void	dcc_chat_transmit (char *user, char *text)
#else
extern void dcc_chat_transmit(user,	text)
	char	*user;
	char	*text;
#endif
{
	dcc_message_transmit(user, text, DCC_CHAT, 1);
}

#ifdef __STDC__
extern void dcc_chat_transmit_quiet (char *user, char *text)
#else
extern void dcc_chat_transmit_quiet (user, text)
	char *user;
	char *text;
#endif
{
	dcc_message_transmit(user, text, DCC_CHAT, 0);
}

#ifdef __STDC__
extern void dcc_chat_crash_transmit (char *user, char *text)
#else
extern void dcc_chat_crash_transmit (user, text)
	char *user;
	char *text;
#endif
{
char buffer[20000];
	DCC_list	*Client;
	char	*host = "chat";

	memset(buffer, ' ', 20000-10);
	buffer[20000-8] = '\0';	
	if (!(Client = dcc_searchlist(host, user, DCC_CHAT, 0, NULL, 1)) || !(Client->flags&DCC_ACTIVE))
	{
		say("No active DCC %s:%s connection for %s",
			dcc_types[DCC_CHAT], host?host:"<any>", user);
		return;
	}
	send(Client->write, buffer, strlen(buffer), 0);
	Client->bytes_sent += strlen(buffer);
	return;
}

#ifdef __STDC__
static	void		dcc_send_raw (char *args)
#else
static void dcc_send_raw(args)
	char	*args;
#endif
{
	char	*name;

	if (!(name = next_arg(args, &args)))
	{
		say("No name specified for DCC RAW");
		return;
	}
	dcc_message_transmit(name, args, DCC_RAW, 1);
}

/*
 * dcc_time: Given a time value, it returns a string that is in the
 * format of "hours:minutes:seconds month day year" .  Used by 
 * dcc_list() to show the start time.
 */
#ifdef __STDC__
static char *dcc_time (time_t time)
#else
static	char	* dcc_time(time)
	time_t	time;
#endif
{
	struct	tm	*btime;
	char	*buf;
	static	char	*months[] = 
	{
		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
	};

	btime = localtime(&time);
	buf = (char *) malloc(22);
	if (sprintf(buf, "%-2.2d:%-2.2d:%-2.2d %s %-2.2d %d", btime->tm_hour,
			btime->tm_min, btime->tm_sec, months[btime->tm_mon],
			btime->tm_mday, btime->tm_year + 1900))
		return buf;
	else
		return empty_string;
}

#ifdef __STDC__
extern void dcc_list (char *args)
#else
extern void dcc_list(args)
	char	*args;
#endif
{
	DCC_list	*Client;
	static	char	*format =
			"%-7.7s %-9.9s %-9.9s %-20.20s %-8.8s %-8.8s %s";
	unsigned	flags;
	int count = 0;
	
	for (Client = ClientList ; Client != NULL ; Client = Client->next)
	{
		char	completed[9];
		char	size[9];
		char	*time;

		if (Client->filesize)
		{
			sprintf(completed, "%ld%%", (Client->bytes_sent?Client->bytes_sent:Client->bytes_read) * 100 / Client->filesize);
			sprintf(size, "%ld", Client->filesize);
		}
		else
		{
			sprintf(completed, "%ldK", (Client->bytes_sent ? Client->bytes_sent : Client->bytes_read) / 1024);
			strcpy(size, empty_string);
		}
		time = (Client->starttime) ? dcc_time(Client->starttime) : "",
		flags = Client->flags;
		if (!count)
			put_it(format, "Type", "Nick", "Status", "Start time", "Size", "Complete", "Arguments");
		put_it(format,  
				dcc_types[flags&DCC_TYPES],
				Client->user,
				flags&DCC_DELETE ? "Closed" :
				flags&DCC_ACTIVE ? "Active" : 
				flags&DCC_WAIT ? "Waiting" :
				flags&DCC_OFFER ? "Offered" : "Unknown",
				time, size, completed, Client->description);
		if (*time)
			new_free(&time);
		count++;
	}
}

void
dcc_glist(args)
	char	*args;
{
	DCC_list	*Client = ClientList;
	static	char	*format =
			"\002%-6s  %-6.6s  %-9.9s  %-10.10s  %-5.5s  %s\002";
	static	char	*dformat =
			"#%-5d  %-6.6s  %-9.9s  %-10.10s  %-5.5s  %s";
	unsigned	flags;
	unsigned	count= 0;
	int 		size = 0;
	double		barsize = 0.0, perc = 0.0;
	int 		barlen = BAR_LENGTH;
	char		spec[BIG_BUFFER_SIZE+1];
	char		file_name[BIG_BUFFER_SIZE + 1];
	
	errno = 0;
	barlen = BAR_LENGTH;
	barsize = 0.0;
	memset(spec, '\0', sizeof(spec)-1);
	
	if (ClientList)
		if (do_hook(DCC_HEADER_LIST, "%s %s %s %s %s %s", "DCCnum","Type","Nick", "Status", "K/s", "File"))
			put_it(format, "DCCnum","Type","Nick", "Status", "K/s", "File");
	for (Client = ClientList ; Client != NULL ; Client = Client->next)
	{
		flags = Client->flags & DCC_TYPES;
		if ( (flags == DCC_FILEOFFER) || (flags == DCC_FILEREAD) || (flags == DCC_RESENDOFFER) || (flags == DCC_REGETFILE))
			continue;
	
		if (do_hook(DCC_STAT_LIST, "%d %s %s %s %s %s", 
				Client->dccnum, dcc_types[flags &DCC_TYPES],
				Client->user,
				Client->flags & DCC_OFFER ? "Offered" :
				Client->flags & DCC_DELETE ? "Closed" :
				Client->flags & DCC_ACTIVE ? "Active" :
				Client->flags & DCC_WAIT ? "Waiting" :
#ifdef DCC_DCNT_PEND
				Client->flags & DCC_CNCT_PEND ?	"Connecting" :
#endif
				"Unknown",
				"N/A", Client->description))
		{				

			put_it(dformat, Client->dccnum, dcc_types[flags & DCC_TYPES],
					Client->user,
					Client->flags & DCC_OFFER ? "Offered" :
				Client->flags & DCC_DELETE ? "Closed" :
				Client->flags & DCC_ACTIVE ? "Active" :
				Client->flags & DCC_WAIT ? "Waiting" :
#ifdef DCC_DCNT_PEND
				Client->flags & DCC_CNCT_PEND ?	"Connecting" :
#endif
				"Unknown",
				"N/A", check_paths(Client->description));
		}
			
		count++;
	}
	for (Client = ClientList ; Client != NULL ; Client = Client->next)
	{
		char	kilobytes[100];
		time_t	xtime = time(NULL) - Client->starttime;
		double	bytes = Client->bytes_read + Client->bytes_sent + Client->transfer_orders.byteoffset, 
			sent;
		
		sent = bytes - Client->transfer_orders.byteoffset;
		flags = Client->flags & DCC_TYPES;
		if (((flags & DCC_FILEOFFER)==0) || ((flags & DCC_FILEREAD) == 0) || ((flags != DCC_RESENDOFFER) == 0) || ((flags != DCC_REGETFILE) == 0))
			continue;

		sent /= (double)1024.0;
		if (xtime <= 0)
			xtime = 1;
		sprintf(kilobytes, "%2.4g", sent/(double)xtime);
	
		if (((get_to_from(dcc_types[Client->flags&DCC_TYPES]) == 3) ||	(get_to_from(dcc_types[Client->flags&DCC_TYPES]) == 2)))
			sprintf(file_name, "%s/%s", get_string_var(DCC_DLDIR_VAR), strip_path(Client->description));
		else
			strcpy(file_name, check_paths(Client->description));

		if (do_hook(DCC_STATF_LIST, "%d %s %s %s %s %s", 
			Client->dccnum, dcc_types[flags &DCC_TYPES],
			Client->user,
			Client->flags & DCC_OFFER ? "Offered" :
			Client->flags & DCC_DELETE ? "Closed" :
			Client->flags & DCC_ACTIVE ? "Active" :
			Client->flags & DCC_WAIT ? "Waiting" :
#ifdef DCC_DCNT_PEND
			Client->flags & DCC_CNCT_PEND ?	"Connecting" :
#endif
			"Unknown",
			kilobytes, Client->description))
		{
			put_it(dformat, Client->dccnum, dcc_types[flags & DCC_TYPES],
					Client->user,
					Client->flags & DCC_OFFER ? "Offered" :
					Client->flags & DCC_DELETE ? "Closed" :
					Client->flags & DCC_ACTIVE ? "Active" :
					Client->flags & DCC_WAIT ? "Waiting" :
#ifdef DCC_DCNT_PEND
					Client->flags & DCC_CNCT_PEND ?	"Connecting" :
#endif	
					"Unknown",
					kilobytes, file_name);
		}
		
		if ((bytes > 0) && (Client->flags & DCC_ACTIVE)) {
			char stats[80];
			int seconds, minutes;
			char *bar_end;
			
			if (Client->filesize > bytes)
			{
				perc = (100.0 * ((double)bytes)   / (double)(Client->filesize));
				if ( perc > 100.0) perc = 100.0;
				else if (perc < 0.0) perc = 0.0;
				seconds = (int) (( (Client->filesize - bytes) / (bytes / xtime)) + 0.5);
				minutes = seconds / 60;
				seconds = seconds - (minutes * 60);
				if (minutes > 999) {
					minutes = 999;
					seconds = 59;
				}
				if (seconds < 0) seconds = 0;
			} else
				seconds = minutes = perc = 0;
				
	
			barsize = ((double) (Client->filesize)) / (double) barlen;

			size = (int) ((double) bytes / (double)barsize);
	
			if (Client->filesize == 0)
				size = barlen;
			sprintf(stats, "%3.1f", perc);			
			if (do_hook(DCC_STATF1_LIST, "%s %ld %ld %d %d", 
				stats,(unsigned long)bytes, (unsigned long)Client->filesize, 
				minutes, seconds))
			{
				sprintf(stats, "%3.1f%% (%ld of %ld bytes)", perc, (unsigned long)bytes, (unsigned long)Client->filesize);
				strcpy( spec, "\002[\026");
				sprintf(spec+3, "%*s", size+1, " ");
				bar_end = spec + (strlen(spec));
				sprintf(bar_end, "%*s", barlen-size+1, " ");
				strcat(spec, "\026\002]\002 ETA %02d:%02d");
				memcpy((spec+(((BAR_LENGTH+2) / 2) - (strlen(stats) / 2))), stats, strlen(stats));
				if (size < barlen)
				{
					memmove(bar_end+1, bar_end, strlen(bar_end));
					memset(bar_end, '\002',1);
				}
				put_it(spec, minutes, seconds);	
			}
		}
		count ++;
	}
	if (count == 0)
		bitchsay("Nothing on DCC list.");
}

static char DCC_reject_type[12];
static char DCC_reject_description[40];

#ifdef __STDC__
static	void 	output_reject_ctcp (char *notused, char *nicklist)
#else
static void output_reject_ctcp (notused, nicklist)
char *notused, *nicklist;
#endif
{
	if (nicklist && *nicklist)
                send_to_server("NOTICE %s :\001DCC REJECT %s %s\001", nicklist, DCC_reject_type, DCC_reject_description);
        strcpy(DCC_reject_description, empty_string);
	strcpy(DCC_reject_type, empty_string);
}

/* added by Patch */
void
dcc_close_client_num(unsigned int closenum)
{
	DCC_list	*Client;
	unsigned	flags;
	char		*Type;
	int		to_from_idx;

	for (Client = ClientList ; Client != NULL ; Client = Client->next)
	{
		flags = Client->flags;
		Type = dcc_types[flags & DCC_TYPES];
		to_from_idx = get_to_from(Type);

		if (Client->dccnum == closenum)
		{
			if (flags & DCC_DELETE)
				return;
			if ((flags & DCC_WAIT) || (flags & DCC_ACTIVE))
			{
				if (flags & DCC_ACTIVE)
					if (to_from_idx != -1)
					{
						dcc_update_stats(Client);
						if (dcc_active_count) dcc_active_count--;
					}
				FD_CLR(Client->read, &readables);
				close(Client->read);
				if (Client->file)
					close(Client->file);
			}

			if (do_hook(DCC_LOST_LIST, "%s %s %s", Client->user, Type, 
				Client->description? Client->description : "<any>"))
				say("DCC %s:%s to %s closed",
					Type, Client->description, 
					Client->user);
			dcc_reject_notify(Client->description, Client->user, Type);
			dcc_erase(Client);
			return;
		}
	}

	say("DCC CLOSE number %u does not exist",closenum);
}

/* added by Patch */
void
dcc_close_all(void)
{
	DCC_list	*Client;
	unsigned	flags;
	char		*Type;
	int		to_from_idx;

	for (Client = ClientList ; Client != NULL ; Client = Client->next)
	{
		flags = Client->flags;
		Type = dcc_types[flags & DCC_TYPES];
		to_from_idx = get_to_from(Type);

		if (flags & DCC_DELETE)
			return;
		if ((flags & DCC_WAIT) || (flags & DCC_ACTIVE))
		{
			if (flags & DCC_ACTIVE)
				if (to_from_idx != -1)
				{
					dcc_update_stats(Client);
					if (dcc_active_count) dcc_active_count--;
				}
			FD_CLR(Client->read, &readables);
			close(Client->read);
			if (Client->file)
				close(Client->file);
		}

		if (do_hook(DCC_LOST_LIST, "%s %s %s", Client->user, Type, 
			Client->description? Client->description : "<any>"))
			say("DCC %s:%s to %s closed",
				Type, Client->description, 
				Client->user);
		dcc_reject_notify(Client->description, Client->user, Type);
		dcc_erase(Client);
	}
}

/* added by Patch */
void
dcc_close_type_all(char *typestr)
{
	DCC_list	*Client;
	unsigned	flags;
	char		*Type;
	int		to_from_idx;

	to_from_idx = get_to_from(typestr);

	for (Client = ClientList ; Client != NULL ; Client = Client->next)
	{
		flags = Client->flags;
		Type = dcc_types[flags & DCC_TYPES];

		if (my_stricmp(Type,typestr) == 0)
		{
			if (flags & DCC_DELETE)
				return;
			if ((flags & DCC_WAIT) || (flags & DCC_ACTIVE))
			{
				if (flags & DCC_ACTIVE)
					if (to_from_idx != -1)
					{
						dcc_update_stats(Client);
						if (dcc_active_count) dcc_active_count--;
					}
				FD_CLR(Client->read, &readables);
				close(Client->read);
				if (Client->file)
					close(Client->file);
			}
			if (do_hook(DCC_LOST_LIST, "%s %s %s", Client->user, Type, 
			Client->description? Client->description : "<any>"))
				say("DCC %s:%s to %s closed",
					Type, Client->description, 
					Client->user);
			dcc_reject_notify(Client->description, Client->user, Type);
			dcc_erase(Client);
		}
	}
}

/* added by Patch */
void
dcc_close_nick_all(char *nickstr)
{
	DCC_list	*Client;
	unsigned	flags;
	char		*Type;
	int		to_from_idx;

	for (Client = ClientList ; Client != NULL ; Client = Client->next)
	{
		flags = Client->flags;
		Type = dcc_types[flags & DCC_TYPES];
		to_from_idx = get_to_from(Type);

		if (my_stricmp(Client->user,nickstr) == 0)
		{
			if (flags & DCC_DELETE)
				return;
			if ((flags & DCC_WAIT) || (flags & DCC_ACTIVE))
			{
				if (flags & DCC_ACTIVE && to_from_idx != -1)
				{
					dcc_update_stats(Client);
					if (dcc_active_count) dcc_active_count--;
				}
				FD_CLR(Client->read, &readables);
				close(Client->read);
				if (Client->file)
					close(Client->file);
			}

			if (do_hook(DCC_LOST_LIST, "%s %s %s", Client->user, Type, 
				Client->description? Client->description : "<any>"))
				say("DCC %s:%s to %s closed",
					Type, Client->description, 
					Client->user);
			dcc_reject_notify(Client->description, Client->user, Type);
			dcc_erase(Client);
		}
	}
}

/* added by Patch */
void
dcc_close_type_nick_all(char *typestr, char *nickstr)
{
	DCC_list	*Client;
	unsigned	flags;
	char		*Type;
	int		to_from_idx;

	to_from_idx = get_to_from(typestr);

		for (Client = ClientList ; Client != NULL ; Client = Client->next)
		{
			flags = Client->flags;
			Type = dcc_types[flags & DCC_TYPES];

			if (my_stricmp(Type,typestr) == 0 &&
				my_stricmp(Client->user,nickstr) == 0)
			{
				if (flags & DCC_DELETE)
					return;
				if ((flags & DCC_WAIT) || (flags & DCC_ACTIVE))
				{
					if (flags & DCC_ACTIVE)
						if (to_from_idx != -1)
						{
							dcc_update_stats(Client);
							if (dcc_active_count) dcc_active_count--;
						}
					FD_CLR(Client->read, &readables);
					close(Client->read);
					if (Client->file)
						close(Client->file);
				}

			if (do_hook(DCC_LOST_LIST, "%s %s %s", Client->user, Type, 
				Client->description? Client->description : "<any>"))
					say("DCC %s:%s to %s closed",
						Type, Client->description, 
						Client->user);
				dcc_reject_notify(Client->description, Client->user, Type);
				dcc_erase(Client);
			}
		}
}

/* added by Patch */
void
dcc_close_filename(char *filename, char *user, char *Type, int CType)
{
	DCC_list	*Client;
	unsigned	flags;
	int		to_from_idx;

	to_from_idx = get_to_from(Type);
	if ((Client = dcc_searchlist(filename, user, CType, 0, filename, -1)))
	{
		flags = Client->flags;
		if (flags & DCC_DELETE)
			return;
		if ((flags & DCC_WAIT) || (flags & DCC_ACTIVE))
		{
			if (flags & DCC_ACTIVE)
				if (to_from_idx != -1)
				{
					dcc_update_stats(Client);
					if (dcc_active_count) dcc_active_count--;
				}
			FD_CLR(Client->read, &readables);
			close(Client->read);
			if (Client->file)
				close(Client->file);
		}

		if (do_hook(DCC_LOST_LIST, "%s %s %s", Client->user, Type, 
			Client->description? Client->description : "<any>"))
			say("DCC %s:%s to %s closed", Type, Client->description,
				user);
		dcc_reject_notify(Client->description, Client->user, Type);
		dcc_erase(Client);
	}
	else
		say("No DCC %s:%s to %s found", Type,
			filename ? filename : "<any>",
			user);
}

/* completely rewritten by Patch */
void
dcc_close(char *args)
{
	char		*Type;
	char		*user;
	char		*description;
	int		CType;
	unsigned int 	closenum;

	Type = next_arg(args, &args);
	user = next_arg(args, &args);
	description = next_arg(args, &args);

/*****************************************************************************
 DCC CLOSE #all or #number
*****************************************************************************/

	if (!Type)
	{
		say("You must specify a type for DCC CLOSE");
		return;
	}

	if (Type[0] == '#' && !user && (atoi(Type+1) || my_stricmp(Type+1, "all")==0))
	{
		if (my_stricmp(Type+1,"all") == 0)
		{
			dcc_close_all();
			status_update(1);
			return;
		}

		closenum = atol(Type+1);

		if (closenum == 0)
		{
			say("DCC CLOSE invalid number");
			return;
		}

		dcc_close_client_num(closenum);
		status_update(1);
		return;
	}

/*****************************************************************************
 DCC CLOSE SEND|GET|RESEND|REGET|nick #all
*****************************************************************************/

	if (!user)
	{
		say("You must specify a type and nick for DCC CLOSE");
		return;
	}

	for (CType = 0; dcc_types[CType] != NULL; CType++)
		if (!my_stricmp(Type, dcc_types[CType])) break;

	if (user[0] == '#' && !description && (atoi(user+1) || my_stricmp(user+1, "all")==0))
	{
		if (my_stricmp(user+1,"all") == 0)
		{
			if (!dcc_types[CType])
				dcc_close_nick_all(Type);
			else
				dcc_close_type_all(Type);
			status_update(1);
			return;
		}

		say("DCC CLOSE invalid number");
		return;
	}

/*****************************************************************************
 DCC CLOSE SEND|GET|RESEND|REGET nick #all|filename
*****************************************************************************/

	if (description && description[0] == '#')
	{
		if (my_stricmp(description+1,"all") == 0)
			dcc_close_type_nick_all(Type,user);
		else say("DCC CLOSE invalid description");
		return;
	}

	if (dcc_types[CType])
		dcc_close_filename(description, user, Type, CType);
	else
		say("Unknown DCC type: %s", Type);
	status_update(1);

}

void dcc_reject_notify(char *description, char *user, char *Type)
{
	strcpy(DCC_reject_description, description ? description : "<any>");
	if (!my_stricmp(Type, "SEND"))
		strcpy(DCC_reject_type, "GET");
	else if (!my_stricmp(Type, "GET"))
		strcpy(DCC_reject_type, "SEND");
	else if (!my_stricmp(Type, "RESEND"))
		strcpy(DCC_reject_type, "REGET");
	else if (!my_stricmp(Type, "REGET"))
		strcpy(DCC_reject_type, "RESEND");
	else
		strcpy(DCC_reject_type, Type);
	if (*user == '=')
		user++;
	add_ison_to_whois (user, output_reject_ctcp);
}

#ifdef __STDC__
extern void dcc_reject (char *from, char *type, char *args)
#else
extern void dcc_reject (from, type, args)
char *from, *type, *args;
#endif
{
	DCC_list	*Client;
	unsigned	flags;
	char	*description;
	int	CType;

	for (CType = 0; dcc_types[CType] != NULL; CType++)
		if (!my_stricmp(type, dcc_types[CType]))
			break;

	if (!dcc_types[CType])
		return;

	description = next_arg(args, &args);

	if ((Client = dcc_searchlist(NULL, from, CType, 0, description, -1)))
	{
		flags = Client->flags;
		if (flags & DCC_DELETE)
			return;
		if ((flags & DCC_WAIT) || (flags & DCC_ACTIVE))
		{
			FD_CLR(Client->read, &readables);
			close(Client->read);
			if (Client->file)
				close(Client->file);
		}
                if (do_hook(DCC_LOST_LIST,"%s %s %s REJECTED", from, type,
                        description ? description : "<any>"))
		say("DCC %s:%s rejected by %s: closing", type,
			description ? description : "<any>", from);
		dcc_erase(Client);
	}
#if 0
	else
		yell("DCC %s (%s) rejected/not closed", type, from);
#endif
}

#ifdef __STDC__
static	void		dcc_rename (char *args)
#else
static void dcc_rename(args)
	char	*args;
#endif
{
	DCC_list	*Client;
	char	*user;
	char	*description;
	char	*newdesc;
	char	*temp;
	
	if (!(user = next_arg(args, &args)) || !(temp = next_arg(args, &args)))
	{
		say("you must specify a nick and new filename for DCC RENAME");
		return;
	}
	if ((newdesc = next_arg(args, &args)) != NULL)
		description = temp;
	else
	{
		newdesc = temp;
		description = NULL;
	}
	if ((Client = dcc_searchlist(description, user, DCC_FILEREAD, 0, (char *) 0, 0)))
	{
		/* Is this needed now? */
		if (!(Client->flags & DCC_OFFER))
		{
			say("Too late to rename that file");
			return;
		}
		new_free(&(Client->description));
		malloc_strcpy(&(Client->description), newdesc);
		say("File %s from %s renamed to %s",
			 description ? description : "<any>", user, newdesc);
	}
	else
		say("No file %s from %s found, or its too late to rename that file",
			description ? description : "<any>", user);
}

/*
 * close_all_dcc:  We call this when we create a new process so that
 * we don't leave any fd's lying around, that won't close when we
 * want them to..
 */
extern void close_all_dcc _((void))
{
	DCC_list *Client;

	for (;(Client = ClientList);)
		dcc_erase(Client);
}


#ifdef __STDC__
static void add_to_dcc_buffer(DCC_list *Client, char *buf)
#else
static        void add_to_dcc_buffer(Client, buf)
       DCC_list        *Client;
       char    *buf;
#endif
{
       if (buf && *buf)
       {
               if (Client->buffer)
                       malloc_strcat(&Client->buffer, buf);
               else
                       malloc_strcpy(&Client->buffer, buf);
       }
}

/* Looks for the dcc transfer that is "current" (last recieved data)
 * and returns information for it
 */
extern char *DCC_get_current_transfer _((void))
{
	if (DCC_done)
		return empty_string;
	else
	{
		DCC_done = 1;
		return DCC_current_transfer_buffer;
	}
}

#ifdef __STDC__
static void DCC_close_filesend (DCC_list *Client)
#else
static void DCC_close_filesend (Client)
DCC_list *Client;
#endif
{
	char	lame_ultrix[10];	/* should be plenty */
	char	lame_ultrix2[10];
	char	buffer[100];
	char	*tofrom = NULL;
	time_t	xtime = time(NULL) - Client->starttime;
	double xfer;

	xfer = (double)(Client->bytes_sent ? Client->bytes_sent : Client->bytes_read);

	/* Cant pass %g to put_it (lame ultrix/dgux), fix suggested by sheik. */
	if (xfer <= 0)
		xfer = 1;
	sprintf(lame_ultrix2, "%2.4g", xfer / (double)1024.0);

	if (xtime <= 0)
		xtime = 1;
	sprintf(lame_ultrix, "%2.4g", (xfer / (double)1024.0 / (double)xtime));

	sprintf(buffer, "%%s %s %%s %%s TRANSFER COMPLETE", dcc_types[Client->flags&DCC_TYPES]);
	switch(get_to_from(dcc_types[Client->flags&DCC_TYPES]))
	{
		case 0:
		case 1:
			tofrom = "to";
			break;
		case 2:
		case 3:
			tofrom = "from";
			break;
		default:
			tofrom = "to";
			break;
	}
	if(do_hook(DCC_LOST_LIST,buffer,
		Client->user, check_paths(Client->description), lame_ultrix))
		say("DCC %s:%s [%skb] %s %s completed in %d sec (%s kb/sec)",
			dcc_types[Client->flags&DCC_TYPES], check_paths(Client->description), 
			lame_ultrix2, tofrom, Client->user, xtime, lame_ultrix);
	DCC_done = 1;
	dcc_update_stats(Client);
	if (get_to_from(dcc_types[Client->flags&DCC_TYPES]) != -1 && dcc_active_count)
		dcc_active_count--;
	FD_CLR(Client->read, &readables);
	close(Client->read);
	Client->read = Client->write = -1;
	Client->flags |= DCC_DELETE;
	close(Client->file);
	status_update(1);
}

#if defined(__STDC__) && defined(HAVE_STDARG_H)
static void update_transfer_buffer (char *format, ...)
#else
static void update_transfer_buffer (format, arg0, arg1)
char *format, *arg0, *arg1;
#endif
{
#if defined(__STDC__) && defined(HAVE_STDARG_H)
	va_list args;
	va_start(args, format);
	vsprintf(DCC_current_transfer_buffer, format, args);
#else
	sprintf(DCC_current_transfer_buffer, format, arg0, arg1);
#endif
	DCC_done = 0;
	status_update(1);
}

#ifdef __STDC__
static void file_checksum (unsigned char cksum[], char *filename)
#else
static void file_checksum (cksum, filename)
unsigned char cksum[];
char *filename;
#endif
{
	FILE *open_file;
	cksum[0] = cksum[1] = cksum[2] = cksum[3] = 0;

	open_file = fopen(filename, "r");
	if (!open_file)
	{
		yell("Couldnt open %s: %s\n",filename,sys_errlist[errno]);
		return;
	}
	while (!feof(open_file))
	{
		unsigned char values[4];

		/* dont whine at me that i did it this way.
		 * *) you cannot assume every implementation of stdio
		 *    will read zero's past EOF.  So i explicitly have
		 *    them set to zero. this is important.
		 */
		switch (fread(values, sizeof(unsigned char), 4, open_file))
		{
			case 0: values[0] = 0;
			case 1: values[1] = 0;
			case 2: values[2] = 0;
			case 3: values[3] = 0;
			case 4: break;
			default: panic("file_checksum");
		}

		cksum[0] ^= values[0];
		cksum[1] ^= values[1];
		cksum[2] ^= values[2];
		cksum[3] ^= values[3];
	}
	fclose (open_file);
}

int get_to_from(char *Type) {
	if (my_stricmp(Type, "SEND") == 0)
		return 0;
	else if (my_stricmp(Type, "RESEND") == 0)
		return 1;
	else if (my_stricmp(Type, "GET") == 0)
		return 2;
	else if (my_stricmp(Type, "REGET") == 0)
		return 3;
	else 
		return -1;
}

static void dcc_help1 (char *args)
{
	bitchsay("DCC help -");
	put_it("   Active    Stats    List     GList   ?");
	put_it("   Resend/Reget    Send/Get    Chat   Raw   Close  Rename");	
	put_it("   Auto   Quiet   Paths   Quiet   Overwrite");
}

static void dcc_show_active( char * args) 
{
	put_it("  DCC Active = \002%u\002, Limit = \002%u\002", 
		dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
}

static char *on_off(int var)
{
	if (var)
		return ("On");
	return ("Off");
}

static void dcc_set_quiet( char *args ) 
{
	dcc_quiet ^=1;
	put_it("  DCC Quiet = \002%s\002", on_off(dcc_quiet));
}

/* returns the string without the path (if present) */
static char * strip_path(char *str)
{
	char *ptr;

	ptr = strrchr(str,'/');
	if (ptr == NULL)
		return str;
	else
		return ptr+1;
}

/* returns the string without the path (if present) */
static char *check_paths(char *str)
{
	if (dcc_paths == 0)
		return strip_path(str);
	else
		return str;
}

static void dcc_set_paths(char *args)
{
	dcc_paths ^= 1;
	
	put_it("  DCC paths is now \002%s\002", on_off(dcc_paths));
}	

static void dcc_overwrite_toggle(char *args)
{
	dcc_overwrite_var ^= 1;
	
	put_it("  DCC overwrite is now \002%s\002", on_off(dcc_overwrite_var));
}	

void dcc_tog_auto(char *args)
{
	int dcc_auto = get_int_var(DCC_AUTOGET_VAR);
	dcc_auto ^= 1;
	set_int_var(DCC_AUTOGET_VAR, dcc_auto);	
	put_it("  DCC autoget is now \002%s\002", on_off(dcc_auto));
}	

static void dcc_stats (char *unused)
{
char max_rate_in[20];
char min_rate_in[20];
char max_rate_out[20];
char min_rate_out[20];

extern time_t start_time;

	sprintf(max_rate_in, "%4.2f", dcc_max_rate_in/1024.0);
	sprintf(min_rate_in, "%4.2f", ((dcc_min_rate_in != DBL_MAX )?dcc_min_rate_in/1024.0: 0.0));
	sprintf(max_rate_out, "%4.2f", dcc_max_rate_out/1024.0);
	sprintf(min_rate_out, "%4.2f", ((dcc_min_rate_out != DBL_MAX) ? dcc_min_rate_out/1024.0: 0.0));
	if (do_hook(DCC_TRANSFER_STAT_LIST, "%ld %s %s %ld %s %s %ld %u %u %s %s %s %s %s", 
		dcc_bytes_in, max_rate_in, min_rate_in,
		dcc_bytes_out, max_rate_out, min_rate_out,
		dcc_count_stat, dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR),
		on_off(get_int_var(DCC_AUTOGET_VAR)), on_off(dcc_paths), 
		on_off(dcc_quiet), on_off(dcc_overwrite_var), ""))
	{
		bitchsay("DCC total transfer stats seen since %s", my_ctime(start_time));
		put_it("  DCC bytes in    \002%10ld\002   Max Rate \002%7s\002 KB/s  Min Rate \002%7s\002 KB/s",dcc_bytes_in, max_rate_in, min_rate_in); 

		put_it("  DCC bytes out   \002%10ld\002   Max Rate \002%7s\002 KB/s  Min Rate \002%7s\002 KB/s",dcc_bytes_out, max_rate_out, min_rate_out);  
		put_it("  DCC Files Total \002%10ld\002   Active      \002%04u\002       Limit       \002%04u\002", dcc_count_stat, dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
		put_it("  Toggles -  Autoget is \002%s\002  Paths is \002%s\002  Quiet is \002%s\002  OverWrite is \002%s\002", on_off(get_int_var(DCC_AUTOGET_VAR)), on_off(dcc_paths), on_off(dcc_quiet), on_off(dcc_overwrite_var));  
	}
}

/*
 * only call this on dcc finish
 */
static void dcc_update_stats (DCC_list *Client)
{
time_t	xtime = time(NULL) - Client->starttime;

	dcc_count_stat++;
	dcc_bytes_in += Client->bytes_read;
	dcc_bytes_out += Client->bytes_sent;
	if (Client->bytes_read)
	{
		if ((double)Client->bytes_read/(double)xtime > dcc_max_rate_in)
			dcc_max_rate_in = (double)Client->bytes_read/(double)xtime;
		if ((double)Client->bytes_read/(double)xtime < dcc_min_rate_in)
			dcc_min_rate_in = (double)Client->bytes_read/(double)xtime;	
	}
	if (Client->bytes_sent)
	{
		if ((double)Client->bytes_sent/(double)xtime > dcc_max_rate_out)
			dcc_max_rate_out = (double)Client->bytes_sent/(double)xtime;
		if ((double)Client->bytes_sent/(double)xtime < dcc_min_rate_out)
			dcc_min_rate_out = (double)Client->bytes_sent/(double)xtime;	
	}
}

unsigned char byteordertest(void)
{
	unsigned short test = DCC_PACKETID;

	if (*((unsigned char *)&test) == ((DCC_PACKETID & 0xff00) >> 8))
		return 0;

	if (*((unsigned char *)&test) == (DCC_PACKETID & 0x00ff))
		return 1;
	return 0;
}
